"""Generate E3FP fingerprints.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
from __future__ import division, print_function
import os
import logging
import argparse
import sys

from python_utilities.scripting import setup_logging
from python_utilities.parallel import make_data_iterator, Parallelizer, \
                                      ALL_PARALLEL_MODES
from python_utilities.io_tools import touch_dir
from e3fp.config.params import read_params, get_default_value, get_value
from e3fp.conformer.util import mol_from_sdf, MolItemName
from e3fp.fingerprint.fprinter import Fingerprinter
from e3fp.fingerprint.db import FingerprintDatabase
import e3fp.fingerprint.fprint as fp

LEVEL_DEF = get_default_value("fingerprinting", "level", int)
RADIUS_MULTIPLIER_DEF = get_default_value("fingerprinting",
                                          "radius_multiplier", float)
FIRST_DEF = get_default_value("fingerprinting", "first", int)
COUNTS_DEF = get_default_value("fingerprinting", "counts", bool)
STEREO_DEF = get_default_value("fingerprinting", "stereo", bool)
INCLUDE_DISCONNECTED_DEF = get_default_value("fingerprinting",
                                             "include_disconnected", bool)
EXCLUDE_FLOATING_DEF = get_default_value("fingerprinting",
                                         "exclude_floating", bool)

OUT_EXT_DEF = ".fp.bz2"
BITS = 2**32


def fprints_dict_from_sdf(sdf_file, **kwargs):
    """Build fingerprints dict for conformers encoded in an SDF file.

    See `fprints_dict_from_mol` for description of arguments.
    """
    try:
        mol = mol_from_sdf(sdf_file)
    except:
        logging.error("Error retrieving mol from {!s}.".format(sdf_file))
        return False
    fprints_dict = fprints_dict_from_mol(mol, **kwargs)
    return fprints_dict


def fprints_dict_from_mol(mol, bits=BITS, level=LEVEL_DEF,
                          radius_multiplier=RADIUS_MULTIPLIER_DEF,
                          first=FIRST_DEF, counts=COUNTS_DEF,
                          stereo=STEREO_DEF,
                          include_disconnected=INCLUDE_DISCONNECTED_DEF,
                          exclude_floating=EXCLUDE_FLOATING_DEF,
                          out_dir_base=None, out_ext=OUT_EXT_DEF,
                          save=False, all_iters=False, overwrite=False):
    """Build a E3FP fingerprint from a mol with at least one conformer.

    Parameters
    ----------
    mol : RDKit Mol
        Input molecule with one or more conformers to be fingerprinted.
    bits : int
        Set number of bits for final folded fingerprint.
    level : int, optional
        Level/maximum number of iterations of E3FP. If -1 is provided, it runs
        until termination, and `all_iters` is set to False.
    radius_multiplier : float, optional
        Radius multiplier for spherical shells.
    first : int, optional
        First `N` number of conformers from file to fingerprint. If -1, all
        are fingerprinted.
    counts : bool, optional
        Instead of bit-based fingerprints. Otherwise, generate count-based
        fingerprints.
    stereo : bool, optional
        Incorporate stereochemistry in fingerprint.
    include_disconnected : bool, optional
        Include disconnected atoms when hashing and for stereo calculations.
        Turn off purely for testing purposes, to make E3FP more like ECFP.
    exclude_floating : bool, optional:
        Mask atoms with no bonds (usually floating ions) from the fingerprint.
        These are often placed arbitrarily and can confound the fingerprint.
    out_dir_base : str, optional
        Basename of out directory to save fingerprints. Iteration number is
        appended.
    out_ext : str, optional
        Extension on fingerprint pickles, used to determine compression level.
    save : bool, optional
        Save fingerprints to directory.
    all_iters : bool, optional
        Save fingerprints from all iterations to file(s).
    overwrite : bool, optional
        Overwrite pre-existing file.

    Deleted Parameters
    ------------------
    sdf_file : str
        SDF file path.
    """
    name = mol.GetProp("_Name")

    if level is None:
        level = -1

    if save:
        filenames = []
        all_files_exist = True
        if level == -1 or not all_iters:
            if level == -1:
                dir_name = "{!s}_complete".format(out_dir_base)
            else:
                dir_name = "{!s}{:d}".format(out_dir_base, level)
            touch_dir(dir_name)
            filenames.append("{:s}/{!s}{!s}".format(dir_name, name, out_ext))
            if not os.path.isfile(filenames[0]):
                all_files_exist = False
        else:
            for i in range(level + 1):
                dir_name = "{:s}{:d}".format(out_dir_base, i)
                touch_dir(dir_name)
                filename = os.path.join(dir_name, "".join[(name, out_ext)])
                filenames.append(filename)
                if not os.path.isfile(filename):
                    all_files_exist = False

        if all_files_exist and not overwrite:
            logging.warning("All fingerprint files for {!s} already exist. "
                            "Skipping.".format(name))
            return {}

    fingerprinter = Fingerprinter(bits=bits, level=level,
                                  radius_multiplier=radius_multiplier,
                                  counts=counts, stereo=stereo,
                                  include_disconnected=include_disconnected,
                                  exclude_floating=exclude_floating)

    try:
        fprints_dict = {}
        logging.info("Generating fingerprints for {!s}.".format(name))
        for j, conf in enumerate(mol.GetConformers()):
            if j == first:
                j -= 1
                break
            fingerprinter.run(conf, mol)
            # fingerprinter.save_substructs_to_db(substruct_db) #PLACEHOLDER
            level_range = range(level + 1)
            if level == -1 or not all_iters:
                level_range = (level, )
            else:
                level_range = range(level + 1)
            for i in level_range:
                fprint = fingerprinter.get_fingerprint_at_level(i)
                fprint.name = MolItemName.from_str(name).to_conf_name(j)
                # if i not in fprints_dict and j != 0:
                #     fprints_dict[i] = fprints_dict[i-1][:j]
                fprints_dict.setdefault(i, []).append(fprint)
        logging.info("Generated {:d} fingerprints for {!s}.".format(j + 1,
                                                                    name))
    except:
        logging.error("Error generating fingerprints for {:s}.".format(name),
                      exc_info=True)
        return {}

    if save:
        if level == -1 or not all_iters:
            fprints = fprints_dict[max(fprints_dict.keys())]
            try:
                fp.savez(filenames[0], *fprints)
                logging.info("Saved fingerprints for {:s}.".format(name))
            except Exception:
                logging.error(
                    "Error saving fingerprints for {:s} to {:s}".format(
                        name, filenames[0]), exc_info=True)
                return {}
        else:
            try:
                for i, fprints in sorted(fprints_dict.items()):
                    fp.savez(filenames[i], *fprints)
                logging.info("Saved fingerprints for {:s}.".format(name))
            except Exception:
                logging.error(
                    "Error saving fingerprints for {:s} to {:s}".format(
                        name, filenames[i]), exc_info=True)
                return {}

    return fprints_dict


def run(sdf_files, bits=BITS, first=FIRST_DEF, level=LEVEL_DEF,
        radius_multiplier=RADIUS_MULTIPLIER_DEF, counts=COUNTS_DEF,
        stereo=STEREO_DEF, include_disconnected=INCLUDE_DISCONNECTED_DEF,
        exclude_floating=EXCLUDE_FLOATING_DEF, params=None,
        out_dir_base=None, out_ext=OUT_EXT_DEF, db_file=None,
        overwrite=False, all_iters=False, log=None, num_proc=None,
        parallel_mode=None, verbose=False):
    """Generate E3FP fingerprints from SDF files."""
    setup_logging(log, verbose=verbose)

    if params is not None:
        params = read_params(params, fill_defaults=True)
        bits = get_value(params, "fingerprinting", "bits", int)
        first = get_value(params, "fingerprinting", "first", int)
        level = get_value(params, "fingerprinting", "level", int)
        radius_multiplier = get_value(params, "fingerprinting",
                                      "radius_multiplier", float)
        counts = get_value(params, "fingerprinting", "counts", bool)
        stereo = get_value(params, "fingerprinting", "stereo", bool)
        include_disconnected = get_value(params, "fingerprinting",
                                         "include_disconnected", bool)
        exclude_floating = get_value(params, "fingerprinting",
                                     "exclude_floating", bool)

    para = Parallelizer(num_proc=num_proc, parallel_mode=parallel_mode)

    if para.rank == 0:
        logging.info("Initializing E3FP generation.")
        logging.info("Getting SDF files")

        if len(sdf_files) == 1 and os.path.isdir(sdf_files[0]):
            from glob import glob
            sdf_files = glob("{:s}/*sdf*".format(sdf_files[0]))

        data_iterator = make_data_iterator(sdf_files)

        logging.info("SDF File Number: {:d}".format(len(sdf_files)))
        if out_dir_base is not None:
            logging.info("Out Directory Basename: {:s}".format(out_dir_base))
            logging.info("Out Extension: {:s}".format(out_ext))
        if db_file is not None:
            logging.info("Database File: {:s}".format(db_file))
        if db_file is None and out_dir_base is None:
            sys.exit('Either `db_file` or `out_dir_base` must be specified.')
        logging.info("Max First Conformers: {:d}".format(first))
        logging.info("Bits: {:d}".format(bits))
        logging.info("Level/Max Iterations: {:d}".format(level))
        logging.info("Shell Radius Multiplier: {:.4g}".format(
            radius_multiplier))
        logging.info("Stereo Mode: {!s}".format(stereo))
        if include_disconnected:
            logging.info("Connected-only mode: on")
        logging.info("Parallel Mode: {!s}".format(para.parallel_mode))
        logging.info("Starting")
    else:
        data_iterator = iter([])

    fp_kwargs = {"first": first,
                 "bits": bits,
                 "level": level,
                 "radius_multiplier": radius_multiplier,
                 "stereo": stereo,
                 "counts": counts,
                 "include_disconnected": include_disconnected,
                 "exclude_floating": exclude_floating,
                 "out_dir_base": out_dir_base,
                 "out_ext": out_ext,
                 "all_iters": all_iters,
                 "overwrite": overwrite,
                 "save": False}
    if out_dir_base is not None:
        fp_kwargs['save'] = True

    run_kwargs = {
        "kwargs": fp_kwargs, "logging_str": "Generated fingerprints for %s",
        "logging_format": lambda x: os.path.basename(x[0]).split(
            os.extsep)[0]}

    results_iter = para.run_gen(fprints_dict_from_sdf, data_iterator,
                                **run_kwargs)

    if db_file is not None:
        fprints = []
        for result, data in results_iter:
            try:
                fprints.extend(result.get(level, result[max(result.keys())]))
            except AttributeError:
                continue
        if len(fprints) > 0:
            db = FingerprintDatabase(fp_type=type(fprints[0]), level=level)
            db.add_fingerprints(fprints)
            db.save(db_file)
            logging.info("Saved fingerprints to {:s}".format(db_file))
    else:
        list(results_iter)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        """Generate E3FP fingerprints from SDF files.""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('sdf_files', nargs='+', type=str,
                        help="""Path to SDF file(s), each with one molecule
                             and multiple conformers.""")
    parser.add_argument('-b', '--bits', type=int, default=BITS,
                        help="""Set number of bits for final folded
                             fingerprint.""")
    parser.add_argument('--first', type=int, default=FIRST_DEF,
                        help="""Set maximum number of first conformers for
                             which to generate fingerprints.""")
    parser.add_argument('-m', '--level', '--max_iterations', type=int,
                        default=LEVEL_DEF,
                        help="""Maximum number of iterations for fingerprint
                             generation. If -1, fingerprinting is run until
                             termination, and `all_iters` is set to False.""")
    parser.add_argument('-r', '--radius_multiplier', '--shell_radius',
                        type=float, default=RADIUS_MULTIPLIER_DEF,
                        help="""Distance to increment shell radius at around
                             each atom, starting at 0.0.""")
    parser.add_argument('--stereo', type=bool, default=STEREO_DEF,
                        help="""Differentiate by stereochemistry.""")
    parser.add_argument('--counts', type=bool, default=COUNTS_DEF,
                        help="""Store counts-based E3FC instead of default
                             bit-based.""")
    parser.add_argument('--params', type=str, default=None,
                        help="""INI formatted file with parameters. If
                             provided, all parameters controlling conformer
                             generation are ignored.""")
    # parser.add_argument('--substruct_db', type=str, default=None,
    #                     help="""Filename to save database mapping identifiers
    #                          to substructures.""")
    # parser.add_argument('--out_format', type=str, default="E3FP",
    #                     choices=["E3FP", "RDKit"],
    #                     help="""Format of saved fingerprint.""")
    parser.add_argument('-o', '--out_dir_base', type=str, default=None,
                        help="""Basename for output directory to save
                             fingerprints. Iteration number is appended to
                             basename.""")
    parser.add_argument('--out_ext', type=str, default=OUT_EXT_DEF,
                        choices=[".fp.pkl", ".fp.gz", ".fp.bz2"],
                        help="""Extension for fingerprint pickles.""")
    parser.add_argument('-d', '--db_file', type=str,
                        default='fingerprints.fps.bz2',
                        help="""Output fingerprint database file.""")
    parser.add_argument('--all_iters', action='store_true',
                        help="""Save fingerprints from all iterations to
                             file(s).""")
    parser.add_argument('-O', '--overwrite', action="store_true",
                        help="""Overwrite existing file(s).""")
    parser.add_argument('-l', '--log', type=str, default=None,
                        help="Log filename.")
    parser.add_argument('-p', '--num_proc', type=int, default=None,
                        help="""Set number of processors to use.""")
    parser.add_argument('--parallel_mode', type=str, default=None,
                        choices=list(ALL_PARALLEL_MODES),
                        help="""Set parallelization mode to use.""")
    parser.add_argument('-v', '--verbose', action="store_true",
                        help="Run with extra verbosity.")
    params = parser.parse_args()

    kwargs = dict(params._get_kwargs())
    sdf_files = kwargs.pop('sdf_files')
    run(sdf_files, **kwargs)



"""Tests for E3FP fingerprinting.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
import glob
import os
import unittest
try:
    import mock
except ImportError:
    import unittest.mock as mock

import numpy as np

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
PLANAR_SDF_FILE = os.path.join(DATA_DIR, "caffeine_planar.sdf.bz2")
NONPLANAR_SDF_FILE = os.path.join(DATA_DIR, "ritalin_nonplanar.sdf.bz2")
RAND_SDF_DIR = os.path.join(DATA_DIR, "rand_sdf_files")
ENANT1_SDF_FILE = os.path.join(DATA_DIR, "stereo1.sdf.bz2")
ENANT2_SDF_FILE = os.path.join(DATA_DIR, "stereo2.sdf.bz2")


class ShellCreationTestCases(unittest.TestCase):

    def test_atom_coords_calculated_correctly(self):
        from e3fp.fingerprint.fprinter import coords_from_atoms
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = [x.GetIdx() for x in mol.GetAtoms()]
        for atom in atoms:
            conf.SetAtomPosition(atom, [0, 0, 0])
        atom_coords = coords_from_atoms(atoms, conf)
        expected_coords = dict(list(zip(atoms, np.zeros((len(atoms), 3),
                                                   dtype=np.float))))
        np.testing.assert_equal(atom_coords, expected_coords)

    def test_distance_matrix_calculated_correctly1(self):
        from e3fp.fingerprint.array_ops import make_distance_matrix
        coords = np.zeros((2, 3), dtype=np.float)
        dist_mat = make_distance_matrix(coords)
        np.testing.assert_almost_equal(
            dist_mat, np.zeros((2, 2), dtype=np.float))

    def test_distance_matrix_calculated_correctly2(self):
        from e3fp.fingerprint.array_ops import make_distance_matrix
        coords = np.array([[0, 0, 0], [-1, -1, -1]], dtype=np.float)
        dist_mat = make_distance_matrix(coords)
        np.testing.assert_almost_equal(
            dist_mat, (3**0.5)*(1 - np.eye(2, dtype=np.float)))

    def test_shells_generator_creation_success(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = [x.GetIdx() for x in mol.GetAtoms()]
        ShellsGenerator(conf, atoms, radius_multiplier=0.5,
                        include_disconnected=True)

    def test_shells_generator_next_works_correctly(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = [x.GetIdx() for x in mol.GetAtoms()]
        shells_gen1 = ShellsGenerator(conf, atoms, radius_multiplier=0.5,
                                      include_disconnected=True)
        shells_gen2 = ShellsGenerator(conf, atoms, radius_multiplier=0.5,
                                      include_disconnected=True)
        self.assertDictEqual(next(shells_gen1), next(shells_gen2))

    def test_connected_match_atoms_rad0_correct(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        shells_gen = ShellsGenerator(conf, atoms, radius_multiplier=0.5,
                                     include_disconnected=True)
        match_atoms = shells_gen.get_match_atoms(0.)
        expect_match_atoms = {k: set() for k in atoms}
        self.assertDictEqual(match_atoms, expect_match_atoms)

    def test_connected_match_atoms_rad1_correct1(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        for atom in atoms:
            conf.SetAtomPosition(atom, [0, 0, 0])
        shells_gen = ShellsGenerator(conf, atoms, radius_multiplier=0.5,
                                     include_disconnected=True)
        match_atoms = shells_gen.get_match_atoms(1.)
        expect_match_atoms = {k: (set(atoms) ^ {k}) for k in atoms}
        self.assertDictEqual(match_atoms, expect_match_atoms)

    def test_connected_match_atoms_rad1_correct2(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        for atom in atoms:
            conf.SetAtomPosition(atom, [0, 0, atom*.75])
        shells_gen = ShellsGenerator(conf, atoms, radius_multiplier=0.5,
                                     include_disconnected=True)
        match_atoms = shells_gen.get_match_atoms(1.)
        expect_match_atoms = {0: {1}, 1: {0, 2}, 2: {1}}
        self.assertDictEqual(match_atoms, expect_match_atoms)

    def test_generates_correct_connected_shells_level0(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        expected_shells_dict = {0: Shell(0), 1: Shell(1), 2: Shell(2)}
        shells_gen = ShellsGenerator(conf, atoms, include_disconnected=False)
        shells_dict = next(shells_gen)
        self.assertDictEqual(shells_dict, expected_shells_dict)

    def test_generates_correct_disconnected_shells_level0(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        expected_shells_dict = {0: Shell(0), 1: Shell(1), 2: Shell(2)}
        shells_gen = ShellsGenerator(conf, atoms)
        shells_dict = next(shells_gen)
        self.assertDictEqual(shells_dict, expected_shells_dict)

    def test_generates_correct_disconnected_shells_level1(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        for atom in atoms:
            conf.SetAtomPosition(atom, [0, 0, .45*atom])
        expected_shells_dict = {0: Shell(0, {1}), 1: Shell(1, {0, 2}),
                                2: Shell(2, {1})}
        shells_gen = ShellsGenerator(conf, atoms, radius_multiplier=0.5,
                                     include_disconnected=True)
        for i in range(2):
            shells_dict = next(shells_gen)
        self.assertDictEqual(shells_dict, expected_shells_dict)

    def test_generates_correct_connected_shells_level1(self):
        from e3fp.fingerprint import fprinter
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        bonds_dict = {0: {1, 2}, 1: {0}, 2: {0}}
        for atom in atoms:
            conf.SetAtomPosition(atom, [0, 0, .45*atom])
        expected_shells_dict = {0: Shell(0, {1}), 1: Shell(1, {0}),
                                2: Shell(2, {})}
        fprinter.bound_atoms_from_mol = mock.Mock(return_value=bonds_dict)

        shells_gen = fprinter.ShellsGenerator(conf, atoms,
                                              radius_multiplier=0.5,
                                              include_disconnected=False)
        for i in range(2):
            shells_dict = next(shells_gen)
        self.assertDictEqual(shells_dict, expected_shells_dict)

    def test_generates_correct_disconnected_shells_level2(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        for atom in atoms:
            conf.SetAtomPosition(atom, [0, 0, .45*atom])
        expected_shells_dict1 = {0: Shell(0, {1}), 1: Shell(1, {0, 2}),
                                 2: Shell(2, {1})}
        expected_shells_dict2 = {
            0: Shell(0, {expected_shells_dict1[1], expected_shells_dict1[2]}),
            1: Shell(1, {expected_shells_dict1[0], expected_shells_dict1[2]}),
            2: Shell(2, {expected_shells_dict1[0], expected_shells_dict1[1]})}
        shells_gen = ShellsGenerator(conf, atoms, radius_multiplier=0.5,
                                     include_disconnected=True)
        for i in range(3):
            shells_dict = next(shells_gen)
        self.assertDictEqual(shells_dict, expected_shells_dict2)

    def test_generates_correct_connected_shells_level2(self):
        from e3fp.fingerprint import fprinter
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        bonds_dict = {0: {1, 2}, 1: {0}, 2: {0}}
        for atom in atoms:
            conf.SetAtomPosition(atom, [0, 0, .45*atom])
        expected_shells_dict1 = {0: Shell(0, {1}), 1: Shell(1, {0}),
                                 2: Shell(2, {})}
        expected_shells_dict2 = {
            0: Shell(0, {expected_shells_dict1[1], expected_shells_dict1[2]}),
            1: Shell(1, {expected_shells_dict1[0]}),
            2: Shell(2, {expected_shells_dict1[0]})}
        fprinter.bound_atoms_from_mol = mock.Mock(return_value=bonds_dict)

        shells_gen = fprinter.ShellsGenerator(conf, atoms,
                                              radius_multiplier=0.5,
                                              include_disconnected=False)
        for i in range(3):
            shells_dict = next(shells_gen)
        self.assertDictEqual(shells_dict, expected_shells_dict2)

    def test_disconnected_substructs_converge(self):
        from e3fp.fingerprint.fprinter import ShellsGenerator
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        for atom in atoms:
            conf.SetAtomPosition(atom, [0, 0, .45*atom])
        shells_gen = ShellsGenerator(conf, atoms, radius_multiplier=0.5,
                                     include_disconnected=True)
        for i in range(4):
            shells_dict = next(shells_gen)
            substructs_dict = {k: v.substruct for k, v
                               in shells_dict.items()}
        next_shells_dict = next(shells_gen)
        next_substructs_dict = {k: v.substruct for k, v
                                in next_shells_dict.items()}

        self.assertDictEqual(substructs_dict, next_substructs_dict)

    def test_connected_substructs_converge(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        atoms = list(range(3))
        bonds_dict = {0: {1, 2}, 1: {0}, 2: {0}}
        for atom in atoms:
            conf.SetAtomPosition(atom, [0, 0, .45*atom])
        fprinter.bound_atoms_from_mol = mock.Mock(return_value=bonds_dict)
        shells_gen = fprinter.ShellsGenerator(conf, atoms,
                                              radius_multiplier=0.5,
                                              include_disconnected=False)
        for i in range(4):
            shells_dict = next(shells_gen)
            substructs_dict = {k: v.substruct for k, v
                               in shells_dict.items()}

        next_shells_dict = next(shells_gen)
        next_substructs_dict = {k: v.substruct for k, v
                                in next_shells_dict.items()}

        self.assertDictEqual(substructs_dict, next_substructs_dict)


class ArrayVectorTestCases(unittest.TestCase):

    def test_rot_matrix_vector_rotation_correct(self):
        from e3fp.fingerprint.array_ops import make_rotation_matrix, \
                                               as_unit
        for i in range(2, 5):
            u0 = as_unit(np.random.uniform(size=3))
            u1 = as_unit(np.random.uniform(size=3))
            rot = make_rotation_matrix(u0, u1)
            np.testing.assert_array_almost_equal(u1, np.dot(rot, u0))

    def test_rot_matrix_array_rotation_correct(self):
        from e3fp.fingerprint.array_ops import make_rotation_matrix, \
                                               as_unit
        for i in range(1, 5):
            u0 = as_unit(np.random.uniform(size=3))
            arr = np.random.uniform(size=(i, 3))
            v = arr[0, :]
            rot = make_rotation_matrix(v, u0)
            rot_arr = np.dot(rot, arr.T).T
            u1 = as_unit(rot_arr[0, :])
            np.testing.assert_array_almost_equal(u0.flatten(), u1.flatten())

    def test_transform_matrix_rotation_only_correct(self):
        from e3fp.fingerprint.array_ops import make_transform_matrix, \
                                               as_unit, pad_array, \
                                               unpad_array, Y_AXIS
        for i in range(1, 5):
            center = np.zeros(3, dtype=np.float)
            arr = np.random.uniform(size=(i, 3)) + center
            y = arr[0, :]
            trans_mat = make_transform_matrix(center, y)
            pad_arr = pad_array(arr)
            rot_arr = unpad_array(np.dot(trans_mat, pad_arr.T).T)
            y0 = as_unit(rot_arr[0, :])
            np.testing.assert_array_almost_equal(y0.flatten(),
                                                 Y_AXIS.flatten())

    def test_transform_matrix_with_translation_correct(self):
        from e3fp.fingerprint.array_ops import make_transform_matrix, \
                                               as_unit, pad_array, \
                                               unpad_array, Y_AXIS
        for i in range(2, 7):
            arr = np.random.uniform(size=(i, 3))
            center = arr[0, :]
            y = arr[1, :] - center
            trans_mat = make_transform_matrix(center, y)
            pad_arr = pad_array(arr)
            rot_arr = unpad_array(np.dot(trans_mat, pad_arr.T).T)
            c0 = rot_arr[0, :]
            y0 = as_unit(rot_arr[1, :])
            np.testing.assert_array_almost_equal(c0.flatten(),
                                                 np.zeros(3))
            np.testing.assert_array_almost_equal(y0.flatten(),
                                                 Y_AXIS.flatten())

    def test_two_axis_transform_correct1(self):
        from e3fp.fingerprint.array_ops import make_transform_matrix, \
                                               as_unit, pad_array, \
                                               unpad_array, Y_AXIS
        for i in range(3, 8):
            arr = np.random.uniform(size=(i, 3))
            center = arr[0, :]
            y = arr[1, :] - center
            z = arr[2, :] - center
            trans_mat = make_transform_matrix(center, y, z)
            pad_arr = pad_array(arr)
            rot_arr = unpad_array(np.dot(trans_mat, pad_arr.T).T)
            c0 = rot_arr[0, :]
            y0 = as_unit(rot_arr[1, :])
            z0 = as_unit(rot_arr[2, :])
            np.testing.assert_array_almost_equal(c0.flatten(),
                                                 np.zeros(3))
            np.testing.assert_array_almost_equal(y0.flatten(),
                                                 Y_AXIS.flatten())
            self.assertAlmostEqual(z0[0], 0.)

    def test_two_axis_transform_correct2(self):
        from e3fp.fingerprint.array_ops import make_transform_matrix, \
                                               as_unit, transform_array, \
                                               Y_AXIS
        for i in range(3, 8):
            arr = np.random.uniform(size=(i, 3))
            center = arr[0, :]
            y = arr[1, :] - center
            z = arr[2, :] - center
            trans_mat = make_transform_matrix(center, y, z)
            rot_arr = transform_array(trans_mat, arr)
            c0 = rot_arr[0, :]
            y0 = as_unit(rot_arr[1, :])
            z0 = as_unit(rot_arr[2, :])
            np.testing.assert_array_almost_equal(c0.flatten(),
                                                 np.zeros(3))
            np.testing.assert_array_almost_equal(y0.flatten(),
                                                 Y_AXIS.flatten())
            self.assertAlmostEqual(z0[0], 0.)


class FingerprinterCreationTestCases(unittest.TestCase):

    def test_main_parameter_ranges_run_without_fail(self):
        from e3fp.fingerprint.fprinter import Fingerprinter
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        conf = mol.GetConformers()[0]
        stereo_opts = (True, False)
        counts_opts = (True, False)
        include_disconnected_opts = (True, False)
        for bits in (1024, 4096, 2**32):
            for level, remove_substructs in [(-1, True), (5, False)]:
                for stereo in stereo_opts:
                    for counts in counts_opts:
                        for include_disconnected in include_disconnected_opts:
                            fprinter = Fingerprinter(
                                bits=bits, level=level, stereo=stereo,
                                counts=counts,
                                remove_duplicate_substructs=remove_substructs,
                                include_disconnected=include_disconnected)
                            fprinter.run(conf, mol)
                            fprinter.get_fingerprint_at_level()

    def test_repeated_runs_produce_same_results(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        level = 2
        conf = mol.GetConformers()[0]
        ref_identifiers = None
        for i in range(5):
            fpr = fprinter.Fingerprinter(level=level, stereo=True,
                                         radius_multiplier=1.718)
            fpr.run(conf, mol)
            identifiers = sorted([x.identifier
                                  for x in fpr.level_shells[level]])
            if ref_identifiers is None:
                ref_identifiers = identifiers
            else:
                self.assertEqual(identifiers, ref_identifiers)

    def test_fingerprint_is_transform_invariant(self):
        from e3fp.fingerprint import fprinter
        from e3fp.fingerprint.array_ops import make_transform_matrix,\
                                               transform_array
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        level = 5
        conf = mol.GetConformers()[0]
        ref_fp = None
        atom_ids = [x.GetIdx() for x in mol.GetAtoms()]
        coords = np.array(list(map(conf.GetAtomPosition, atom_ids)),
                          dtype=np.float)
        for i in range(5):
            rand_y = np.random.uniform(size=3)
            rand_trans = np.random.uniform(size=3)*100
            trans_mat = make_transform_matrix(rand_trans)
            rot_mat = make_transform_matrix(np.zeros(3), rand_y)
            transform_mat = np.dot(trans_mat, rot_mat)
            new_coords = transform_array(transform_mat, coords)
            with self.assertRaises(AssertionError):
                np.testing.assert_almost_equal(new_coords, coords)

            for atom_id, new_coord in zip(atom_ids, new_coords):
                conf.SetAtomPosition(atom_id, new_coord)
            test_coords = np.array(list(map(conf.GetAtomPosition, atom_ids)),
                                   dtype=np.float)
            np.testing.assert_almost_equal(test_coords, new_coords)

            fpr = fprinter.Fingerprinter(level=level, stereo=True,
                                         radius_multiplier=1.718)
            fpr.run(conf, mol)
            fp = fpr.get_fingerprint_at_level(level)
            if ref_fp is None:
                ref_fp = fp
            else:
                self.assertEqual(fp, ref_fp)

    def test_quick(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        level = 5
        conf = mol.GetConformers()[0]
        fpr = fprinter.Fingerprinter(level=level, bits=1024, stereo=True,
                                      radius_multiplier=1.718)
        fpr.run(conf, mol)
        # fprint = fpr.get_fingerprint_at_level(level)
        # fpr.substructs_to_pdb(reorient=True)


class HashingTestCases(unittest.TestCase):
    def test_hashing_wrong_dtype_fails1(self):
        from e3fp.fingerprint.fprinter import hash_int64_array
        with self.assertRaises(TypeError):
            hash_int64_array(np.arange(3, dtype=np.float))

    def test_hashing_wrong_dtype_fails2(self):
        from e3fp.fingerprint.fprinter import hash_int64_array
        with self.assertRaises(TypeError):
            hash_int64_array(np.arange(3, dtype=np.int32))

    def test_hashing_produces_example_result(self):
        from e3fp.fingerprint.fprinter import hash_int64_array
        self.assertEqual(hash_int64_array(np.array([42], dtype=np.int64)),
                          1871679806)

    def test_mmh3_produces_example_result(self):
        import mmh3
        self.assertEqual(mmh3.hash("foo", 0), -156908512)


class StereoTestCases(unittest.TestCase):

    def test_stereo_indicators_for_frame(self):
        from e3fp.fingerprint.fprinter import stereo_indicators_from_shell
        from e3fp.fingerprint.array_ops import project_to_plane,\
                                               make_transform_matrix,\
                                               transform_array
        from e3fp.fingerprint.structs import Shell

        shell = Shell(0, {1, 2, 3})
        atom_coords = np.asarray([[0,  0,   0.],
                                  [1, -0.5, 0.],
                                  [0,  2.,  0.],  # -> y
                                  [0,  0,   3.]],  # -> z
                                 dtype=np.float)
        atom_tuples = [(1, 1, Shell(2)),  # -> y
                       (2, 1, Shell(1)),
                       (5, 5, Shell(3))]  # -> z
        for i in range(20):
            rand_trans = np.random.uniform(size=3)*100
            rand_y = np.random.uniform(size=3)*10
            rand_v = np.random.uniform(size=3)*20
            rand_z = project_to_plane(rand_v, rand_y)*30
            rand_transform_mat = make_transform_matrix(np.zeros(3), rand_y,
                                                       rand_z)
            new_coords = transform_array(rand_transform_mat, atom_coords)
            np.testing.assert_almost_equal(
                atom_coords, transform_array(
                    np.linalg.inv(rand_transform_mat), new_coords))
            new_coords += rand_trans

            atom_coords_dict = dict(list(zip(list(range(4)), new_coords)))
            stereo_ind = stereo_indicators_from_shell(shell, atom_tuples,
                                                      atom_coords_dict)
            # 2 is chosen for y, 3 for z
            expect_stereo_ind = [1, -5, 2]
            self.assertEqual(stereo_ind, expect_stereo_ind)

    def test_empty_tuples_returns_empty(self):
        from e3fp.fingerprint.fprinter import stereo_indicators_from_shell
        from e3fp.fingerprint.structs import Shell

        shell = Shell(0, {})
        atom_coords_dict = {0: np.random.uniform(size=3)}
        atom_tuples = []
        stereo_ind = stereo_indicators_from_shell(shell, atom_tuples,
                                                  atom_coords_dict)
        self.assertEqual(len(stereo_ind), 0)

    def test_no_unique_y_two_evenly_spaced_correct(self):
        from e3fp.fingerprint.fprinter import stereo_indicators_from_shell
        from e3fp.fingerprint.structs import Shell
        shell = Shell(0, {1, 2})
        atom_coords_dict = {0: [0, 0, 0], 1: [0, 1, 0], 2: [0, 0, 1]}
        atom_tuples = [(1, 1, Shell(1)),
                       (1, 1, Shell(2))]
        stereo_ind = stereo_indicators_from_shell(shell, atom_tuples,
                                                  atom_coords_dict)
        # mean should be between 1 and 2, so z cannot be picked, so all 0.
        expect_stereo_ind = [1, 2]

        self.assertEqual(stereo_ind, expect_stereo_ind)

    def test_no_unique_y_three_evenly_spaced_produces_zeros(self):
        from e3fp.fingerprint.fprinter import stereo_indicators_from_shell
        from e3fp.fingerprint.structs import Shell
        shell = Shell(0, {1, 2, 3})
        atom_coords_dict = {0: [0, 0, 0], 1: [0, 1, 0], 2: [0, 0, 1],
                            3: [1, 0, 0]}
        atom_tuples = [(1, 1, Shell(1)),
                       (1, 1, Shell(2)),
                       (1, 1, Shell(3))]
        stereo_ind = stereo_indicators_from_shell(shell, atom_tuples,
                                                  atom_coords_dict)
        # mean should be between 1 and 2, so z cannot be picked, so all 0.
        expect_stereo_ind = [0, 0, 0]

        self.assertEqual(stereo_ind, expect_stereo_ind)

    def test_no_unique_y_along_poles_correct(self):
        from e3fp.fingerprint.fprinter import stereo_indicators_from_shell
        from e3fp.fingerprint.structs import Shell
        shell = Shell(0, {1, 2, 3})
        atom_coords_dict = {0: [0, 0, 0], 1: [0, 1, 0], 2: [0, -5, 0],
                            3: [0, -10, 0]}
        atom_tuples = [(1, 1, Shell(1)),
                       (1, 1, Shell(2)),
                       (1, 1, Shell(3))]
        stereo_ind = stereo_indicators_from_shell(shell, atom_tuples,
                                                  atom_coords_dict)
        # mean should be along atom 2/3, so z not be picked, but all at poles.
        expect_stereo_ind = [-1, 1, 1]

        self.assertEqual(stereo_ind, expect_stereo_ind)

    def test_all_indicators_correctly_assigned(self):
        from e3fp.fingerprint.fprinter import quad_indicators_from_coords
        cent_coords = np.array([[ 0.,  1.,  1.],
                                [-1.,  1.,  0.],
                                [ 0.,  1., -1.],
                                [ 1.,  1.,  0.],
                                [ 0., -1.,  1.],
                                [-1., -1.,  0.],
                                [ 0., -1., -1.],
                                [ 1., -1.,  0.]], dtype=np.float)
        y = np.array([0, 1, 0], dtype=np.float)
        y_ind = None
        z = np.array([0, 0, 1], dtype=np.float)
        long_sign = np.array([1, 1, 1, 1, -1, -1, -1, -1], dtype=np.int64)
        quad = quad_indicators_from_coords(cent_coords, y, y_ind, z,
                                           long_sign)
        expect_quad = [2, 3, 4, 5, -2, -3, -4, -5]
        self.assertEqual(list(quad), expect_quad)

    def test_stereo_sets_correct_transform_matrix(self):
        from e3fp.fingerprint.fprinter import stereo_indicators_from_shell
        from e3fp.fingerprint.array_ops import project_to_plane, \
                                               make_transform_matrix, \
                                               transform_array
        from e3fp.fingerprint.structs import Shell

        shell = Shell(0, {1, 2})
        atom_coords = np.asarray([[0,  0,   0.],
                                  [0,  2.,  0.],  # -> y
                                  [0,  0,   3.]],  # -> z
                                 dtype=np.float)
        atom_tuples = [(1, 1, Shell(1)),  # -> y
                       (5, 5, Shell(2))]  # -> z
        for i in range(20):
            rand_trans = np.random.uniform(size=3)*100
            rand_y = np.random.uniform(size=3)*10
            rand_v = np.random.uniform(size=3)*20
            rand_z = project_to_plane(rand_v, rand_y)*30
            rand_transform_mat = make_transform_matrix(np.zeros(3), rand_y,
                                                       rand_z)
            trans_mat = np.identity(4, dtype=np.float)
            trans_mat[:3, 3] = rand_trans
            rand_transform_mat = np.dot(trans_mat, rand_transform_mat)

            new_coords = transform_array(rand_transform_mat, atom_coords)
            reverse_trans_mat = np.linalg.inv(rand_transform_mat)

            np.testing.assert_almost_equal(
                atom_coords, transform_array(reverse_trans_mat, new_coords))

            atom_coords_dict = dict(list(zip(list(range(3)), new_coords)))
            stereo_indicators_from_shell(shell, atom_tuples, atom_coords_dict)
            np.testing.assert_almost_equal(shell.transform_matrix,
                                           reverse_trans_mat)

    def test_no_neighbors_transform_matrix_is_translate(self):
        from e3fp.fingerprint.fprinter import stereo_indicators_from_shell
        from e3fp.fingerprint.structs import Shell
        shell = Shell(0)
        center_coord = np.random.uniform(3)
        stereo_indicators_from_shell(shell, [], {0: center_coord})
        np.testing.assert_almost_equal(shell.transform_matrix[:3, 3],
                                       -center_coord)

    # def test_stereo_planar_molecule_successful(self):
    #     pass

    # def test_more_uniqueness_usecases(self):
    #     pass


class GenerateFingerprintTestCases(unittest.TestCase):

    def test_runs_without_exception_on_random_mols(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        rand_sdf_files = glob.glob(RAND_SDF_DIR+"/*.sdf*")
        rand_sdf_files = [rand_sdf_files[i] for i
                          in np.random.randint(len(rand_sdf_files), size=10)]
        level = 5
        for sdf_file in rand_sdf_files:
            mol = mol_from_sdf(sdf_file)
            conf = mol.GetConformers()[0]
            fpr = fprinter.Fingerprinter(level=level, bits=1024, stereo=True,
                                         radius_multiplier=1.718)
            fpr.run(conf, mol)
            fpr.get_fingerprint_at_level(level)

    def test_initial_identifiers_assigned_correctly(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        level = 0
        conf = mol.GetConformers()[0]
        fpr = fprinter.Fingerprinter(level=level, bits=1024, stereo=True,
                                     radius_multiplier=1.718)
        fpr.run(conf, mol)
        fprint = fpr.get_fingerprint_at_level(0)
        expect_ident = set([48, 124, 185, 484, 617, 674])
        self.assertEqual(set(fprint.indices), expect_ident)

    def test_remove_dupe_substructs_makes_same_substruts_diff_shells(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        level = 2
        conf = mol.GetConformers()[0]
        fpr = fprinter.Fingerprinter(level=level, bits=1024, stereo=True,
                                     radius_multiplier=1.718,
                                     remove_duplicate_substructs=True)
        fpr.run(conf, mol)
        shells_no_dupes = set(fpr.level_shells[fpr.current_level])
        substructs_no_dupes = set([x.substruct for x in shells_no_dupes])

        fpr = fprinter.Fingerprinter(level=level, bits=1024, stereo=True,
                                     radius_multiplier=1.718,
                                     remove_duplicate_substructs=False)
        fpr.run(conf, mol)
        shells_with_dupes = set(fpr.level_shells[fpr.current_level])
        substructs_with_dupes = set([x.substruct for x in shells_with_dupes])

        self.assertEqual(substructs_no_dupes, substructs_with_dupes)
        self.assertNotEqual(shells_no_dupes, shells_with_dupes)

    def test_stereoisomers_produce_nonequal_fingerprints_stereo(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        mol1 = mol_from_sdf(ENANT1_SDF_FILE)
        mol2 = mol_from_sdf(ENANT2_SDF_FILE)
        level = 5
        fpr = fprinter.Fingerprinter(level=level, bits=1024, stereo=True,
                                     radius_multiplier=1.718,
                                     remove_duplicate_substructs=True)
        fpr.run(conf=0, mol=mol1)
        fp1 = fpr.get_fingerprint_at_level(level)
        fpr.run(conf=0, mol=mol2)
        fp2 = fpr.get_fingerprint_at_level(level)
        self.assertNotEqual(fp1, fp2)

    def test_stereoisomers_produce_equal_fingerprints_nonstereo(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        mol1 = mol_from_sdf(ENANT1_SDF_FILE)
        mol2 = mol_from_sdf(ENANT2_SDF_FILE)
        level = 5
        fpr = fprinter.Fingerprinter(level=level, stereo=False,
                                     radius_multiplier=1.718,
                                     remove_duplicate_substructs=True)
        fpr.run(conf=0, mol=mol1)
        fp1 = fpr.get_fingerprint_at_level(level)
        fpr.run(conf=0, mol=mol2)
        fp2 = fpr.get_fingerprint_at_level(level)
        self.assertEqual(fp1, fp2)

    def test_reordering_conformers_produces_same_fprints(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        import random
        rand_sdf_files = glob.glob(os.path.join(RAND_SDF_DIR, "*.sdf*"))
        mol = mol_from_sdf(rand_sdf_files[0])
        level = 5
        fpr = fprinter.Fingerprinter(level=level, stereo=False,
                                     radius_multiplier=1.718,
                                     remove_duplicate_substructs=True)
        conf_ids1 = [x.GetId() for x in mol.GetConformers()]
        fprints1 = {}
        for conf_id in conf_ids1:
            fpr.run(conf_id, mol)
            fprints1[conf_id] = fpr.get_fingerprint_at_level(level)

        conf_ids2 = list(conf_ids1)
        random.shuffle(conf_ids2)
        fprints2 = {}
        for conf_id in conf_ids2:
            fpr.run(conf_id, mol)
            fprints2[conf_id] = fpr.get_fingerprint_at_level(level)
        self.assertEqual(fprints1, fprints2)

    def test_reordering_mols_produces_same_fprints(self):
        from e3fp.fingerprint import fprinter
        from e3fp.conformer.util import mol_from_sdf
        import random
        rand_sdf_files = glob.glob(RAND_SDF_DIR+"/*.sdf*")
        mols = list(map(mol_from_sdf, rand_sdf_files[:5]))
        level = 5
        fpr = fprinter.Fingerprinter(level=level, stereo=False,
                                     radius_multiplier=1.718,
                                     remove_duplicate_substructs=True)
        fprints1 = {}
        for mol in mols:
            fpr.run(conf=0, mol=mol)
            fprints1[mol] = fpr.get_fingerprint_at_level(level)

        random.shuffle(mols)
        fpr = fprinter.Fingerprinter(level=level, stereo=False,
                                     radius_multiplier=1.718,
                                     remove_duplicate_substructs=True)
        fprints2 = {}
        for mol in mols:
            fpr.run(conf=0, mol=mol)
            fprints2[mol] = fpr.get_fingerprint_at_level(level)

        self.assertEqual(fprints1, fprints2)


if __name__ == "__main__":
    unittest.main()



"""Various array operations.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
import numpy as np
from scipy.spatial.distance import pdist, squareform

QUATERNION_DTYPE = np.float64
X_AXIS, Y_AXIS, Z_AXIS = np.identity(3, dtype=np.float64)
EPS = 1e-12  # epsilon, a number close to 0


# Vector Algebra Methods
def as_unit(v, axis=1):
    """Return array of unit vectors parallel to vectors in `v`.

    Parameters
    ----------
    v : ndarray of float
    axis : int, optional
        Axis along which to normalize length.

    Returns
    -------
    ndarray of float : Unit vector of `v`, i.e. `v` divided by its
                       magnitude along `axis`.
    """
    u = np.array(v, dtype=np.float64, copy=True)
    if u.ndim == 1:
        mag = np.sqrt(np.dot(u, u))
        if mag < EPS:
            mag = 1.
    else:
        mag = np.atleast_1d(np.sum(u*u, axis))
        np.sqrt(mag, mag)
        if axis is not None:
            mag = np.expand_dims(mag, axis)
        mag[mag < EPS] = 1.
    u /= mag
    return u


def make_distance_matrix(coords):
    """Build pairwise distance matrix from coordinates.

    Parameters
    ----------
    coords : ndarray of float
        an Mx3 array of cartesian coordinates.

    Returns
    -------
    ndarray of float : square symmetrical distance matrix
    """
    return squareform(pdist(coords))


def make_transform_matrix(center, y=None, z=None):
    """Make 4x4 homogenous transformation matrix.

    Given Nx4 array A where A[:, 4] = 1., the transform matrix M should be
    used with dot(M, A.T).T. Order of operations is 1. translation, 2. align
    `y` x `z` plane to yz-plane 3. align `y` to y-axis.

    Parameters
    ----------
    center : 1x3 array of float
        Coordinate that should be centered after transformation.
    y : None or 1x3 array of float
        Vector that should lie on the y-axis after transformation
    z : None or 1x3 array of float
        Vector that after transformation should lie on yz-plane in direction
        of z-axis.

    Returns
    -------
    4x4 array of float
        4x4 homogenous transformation matrix.
    """
    translate = np.identity(4, dtype=np.float64)
    translate[:3, 3] = -np.asarray(center, dtype=np.float64)
    if y is not None:
        y = np.atleast_2d(y)
        if z is None:
            rotate = np.identity(4, dtype=np.float64)
            rotate[:3, :3] = make_rotation_matrix(y, Y_AXIS)
        else:
            z = np.atleast_2d(z)
            rotate_norm = np.identity(4, dtype=np.float64)
            x_unit = as_unit(np.cross(y, z))
            rotate_norm[:3, :3] = make_rotation_matrix(x_unit, X_AXIS)
            new_y = np.dot(rotate_norm[:3, :3], y.flatten())
            rotate_y = np.identity(4, dtype=np.float64)
            rotate_y[:3, :3] = make_rotation_matrix(new_y.flatten(), Y_AXIS)
            rotate = np.dot(rotate_y, rotate_norm)
        transform = np.dot(rotate, translate)
    else:
        transform = translate
    return transform


def make_rotation_matrix(v0, v1):
    """Create 3x3 matrix of rotation from `v0` onto `v1`.

    Should be used by dot(R, v0.T).T.

    Parameters
    ----------
    v0 : 1x3 array of float
        Initial vector before alignment.
    v1 : 1x3 array of float
        Vector to which to align `v0`.
    """
    v0 = as_unit(v0)
    v1 = as_unit(v1)
    o = np.cross(v0, v1).flatten()
    if np.all(o == 0.):
        return np.identity(3, dtype=np.float64)
    u = as_unit(o).flatten()
    sin_ang = np.linalg.norm(o)
    cos_ang = np.dot(v0, v1.T)
    ux = np.array([[   0., -u[2],  u[1]],
                   [ u[2],    0., -u[0]],
                   [-u[1],  u[0],    0.]], dtype=np.float64)
    rot = (cos_ang * np.identity(3, dtype=np.float64) + sin_ang * ux +
           (1 - cos_ang) * np.outer(u, u))
    return rot


def transform_array(transform_matrix, a):
    """Pad an array with 1s, transform, and return with original dimensions.

    Parameters
    ----------
    transform_matrix : 4x4 array of float
        4x4 homogenous transformation matrix
    a : Nx3 array of float
        Array of 3-D coordinates.

    Returns
    -------
    Nx3 array of float : Transformed array
    """
    return unpad_array(np.dot(transform_matrix, pad_array(a).T).T)


def pad_array(a, n=1., axis=1):
    """Return `a` with row of `n` appended to `axis`.

    Parameters
    ----------
    a : ndarray
        Array to pad
    n : float or int, optional
        Value to pad `a` with
    axis : int, optional
        Axis of `a` to pad with `n`.

    Returns
    -------
    ndarray
        Padded array.
    """
    if a.ndim == 1:
        pad = np.ones(a.shape[0]+1, dtype=a.dtype) * n
        pad[:a.shape[0]] = a
    else:
        shape = list(a.shape)
        shape[axis] += 1
        pad = np.ones(shape, dtype=a.dtype)
        pad[:a.shape[0], :a.shape[1]] = a
    return pad


def unpad_array(a, axis=1):
    """Return `a` with row removed along `axis`

    Parameters
    ----------
    a : ndarray
        Array from which to remove row
    axis : int, optional
        Axis from which to remove row

    Returns
    -------
    ndarray
        Unpadded array.
    """
    if a.ndim == 1:
        return a[:-1]
    else:
        shape = list(a.shape)
        shape[axis] -= 1
        return a[:shape[0], :shape[1]]


def project_to_plane(vec_arr, norm):
    """Project array of vectors to plane with normal `norm`.

    Parameters
    ----------
    vec_arr : Nx3 array
        Array of N 3D vectors.
    norm : 1x3 array
        Normal vector to plane.

    Returns
    -------
    Nx3 array
        Array of vectors projected onto plane.
    """
    unit_norm = as_unit(norm).flatten()
    mag_on_norm = np.dot(vec_arr, unit_norm)
    if vec_arr.ndim == 1:
        vec_on_norm = np.array(unit_norm, copy=True)
        vec_on_norm *= mag_on_norm
    else:
        vec_on_norm = np.tile(unit_norm, (vec_arr.shape[0], 1))
        vec_on_norm *= mag_on_norm[:, None]
    return vec_arr - vec_on_norm


def calculate_angles(vec_arr, ref, ref_norm=None):
    """Calculate angles between vectors in `vec_arr` and `ref` vector.

    If `ref_norm` is not provided, angle ranges between 0 and pi. If it is
    provided, angle ranges between 0 and 2pi. Note that if `ref_norm` is
    orthogonal to `vec_arr` and `ref`, then the angle is rotation around the
    axis, but if a non-orthogonal axis is provided, this may not be the case.

    Parameters
    ----------
    vec_arr : Nx3 array of float
        Array of N 3D vectors.
    ref : 1x3 array of float
        Reference vector
    ref_norm : 1x3 array of float
        Normal vector.

    Returns
    -------
    1-D array
        Array of N angles
    """
    unit_vec_arr = as_unit(vec_arr)
    unit_ref = as_unit(ref).flatten()
    ang = np.arccos(np.clip(np.dot(unit_vec_arr, unit_ref), -1.0, 1.0))
    # handle cases where a vector is the origin
    ang[np.all(unit_vec_arr == np.zeros(3), axis=1)] = 0.
    if ref_norm is not None:
        sign = np.sign(np.dot(ref_norm,
                              np.cross(unit_vec_arr, unit_ref).T)).flatten()
        sign[sign == 0] = 1
        ang = rotate_angles(sign * ang, 2 * np.pi)
    return ang


def rotate_angles(angles, amount):
    """Rotate angles by `amount`, keeping in 0 to 2pi range.

    Parameters
    ----------
    angles : 1-D array of float
        Angles in radians
    amount : float
        Amount to rotate angles by

    Returns
    -------
    1-D array of float : Rotated angles
    """
    return (angles + amount) % (2 * np.pi)


def quaternion_to_transform_matrix(quaternion, translation=np.zeros(3)):
    """Convert quaternion to homogenous 4x4 transform matrix.

    Parameters
    ----------
    quaternion : 4x1 array of float
        Quaternion describing rotation after translation.
    translation : 3x1 array of float, optional
        Translation to be performed before rotation.
    """
    q = np.array(quaternion, dtype=np.float64, copy=True)
    n = np.linalg.norm(q)
    if n < 1e-12:
        return np.identity(4, dtype=np.float64)
    q /= n
    q = 2 * np.outer(q, q)
    transform_mat = np.array([
        [1.-q[2, 2]-q[3, 3],    q[1, 2]-q[3, 0],    q[1, 3]+q[2, 0], 0.],
        [   q[1, 2]+q[3, 0], 1.-q[1, 1]-q[3, 3],    q[2, 3]-q[1, 0], 0.],
        [   q[1, 3]-q[2, 0],    q[2, 3]+q[1, 0], 1.-q[1, 1]-q[2, 2], 0.],
        [                0.,                 0.,                 0., 1.]],
        dtype=np.float64)
    transform_mat[:3, 3] = translation
    return transform_mat


def transform_matrix_to_quaternion(transform_matrix, dtype=QUATERNION_DTYPE):
    """Convert homogenous 4x4 transform matrix to quaternion.

    Parameters
    ----------
    transform_matrix : 4x4 array of float
        Homogenous transformation matrix.
    dtype : numpy dtype, optional
        Datatype for returned quaternion.
    """
    T = np.array(transform_matrix, dtype=np.float64)
    R = T[:3, :3]
    q = np.zeros(4, dtype=dtype)
    q[0] = np.sqrt(1. + R.trace()) / 2.
    q[1] = R[2, 1] - R[1, 2]
    q[2] = R[0, 2] - R[2, 0]
    q[3] = R[1, 0] - R[0, 1]
    q[1:4] /= 4. * q[0]
    return q



"""Tests for loading config files.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
import os
import unittest


class ConfigTestCases(unittest.TestCase):

    def test_config_file_exists(self):
        from e3fp.config.params import DEF_PARAM_FILE
        self.assertTrue(os.path.isfile(DEF_PARAM_FILE))


if __name__ == "__main__":
    unittest.main()



"""Database for storing and serializing fingerprints.

Author: Seth Axen
E-mail: seth.axen@gmail.com"""
from collections import defaultdict
try:
    import cPickle as pkl
except ImportError:  # Python 3
    import pickle as pkl
import logging

from scipy.sparse import vstack
from python_utilities.io_tools import smart_open
from .fprint import Fingerprint, CountFingerprint, FloatFingerprint


class FingerprintDatabase(object):

    """A database for storing, saving, and loading fingerprints.

    Fingerprints must have the same bit length and be of the same level.
    Additionally, they are all be cast to the type of fingerprint passed
    to the database upon instantiation.

    Attributes
    ----------
    array : csr_matrix
        Sparse matrix with dimensions N x M, where M is the number
        of bits in the fingerprints.
    fp_names : list of str
        Names of fingerprints
    fp_names_to_indices : dict
        Map from fingerprint name to row indices of `array`
    fp_type : type
        Type of fingerprint (Fingerprint, CountsFingerprint, FloatFingerprint)
    fp_num : int
        Number of fingerprints in database
    bits : int
        Number of bits of fingerprints
    level : int
        Level, or number of iterations used during fingerprinting.
    name : str
        Name of database
    """

    def __init__(self, fp_type=Fingerprint, level=-1, name=None):
        """Constructor

        Parameters
        ----------
        fp_type : type, optional
            Type of fingerprint (Fingerprint, CountsFingerprint,
            FloatFingerprint)
        level : int, optional
            Level, or number of iterations used during fingerprinting.
        name : str, optional
        Name of database
        """
        if fp_type not in (Fingerprint, CountFingerprint,
                           FloatFingerprint):
            raise TypeError(
                "{} is not a valid fingerprint type".format(fp_type))
        self.name = name
        self.fp_type = fp_type
        self.level = level
        self.array = None
        self.fp_names = []
        self.fp_names_to_indices = defaultdict(list)

    def add_fingerprints(self, fprints):
        """Add fingerprints to database.

        Parameters
        ----------
        fprints : iterable of Fingerprint
            Fingerprints to add to database
        """
        self._check_fingerprints_are_valid(fprints)

        dtype = self.fp_type.vector_dtype

        new_rows, new_names = list(zip(*[
            (fprint.to_vector(sparse=True, dtype=dtype), fprint.name)
            for fprint in fprints]))
        try:
            old_fp_num = self.array.shape[0]
            self.array = vstack([self.array] + list(new_rows))
        except (AttributeError, ValueError):  # array not yet defined
            old_fp_num = 0
            self.array = vstack(new_rows)
        self.array = self.array.tocsr()
        del new_rows
        self.fp_names += new_names
        self.update_names_map(new_names=new_names, offset=old_fp_num)

    def update_names_map(self, new_names=None, offset=0):
        """Update map of fingerprint names to row indices of `self.array`.

        Parameters
        ----------
        new_names : iterable of str, optional
            Names to add to map. If None, map is completely rebuilt.
        offset : int, optional
            Number of rows before new rows.
        """
        if new_names is None:
            new_names = self.fp_names
        for i, name in enumerate(new_names):
            self.fp_names_to_indices[name].append(i + offset)

    def save(self, fn="fingerprints.fps.bz2"):
        """Save database to file.

        Parameters
        ----------
        fn : str, optional
            Filename or basename if extension does not include '.fps'
        """
        if ".fps" not in fn:
            fn += ".fps.bz2"
        with smart_open(fn, "w") as f:
            pkl.dump(self, f)

    @classmethod
    def load(cls, fn):
        """Load database from file.

        Parameters
        ----------
        fn : str
            Filename

        Returns
        -------
        FingerprintDatabase
            Dabatase
        """
        with smart_open(fn) as f:
            return pkl.load(f)

    @property
    def fp_num(self):
        try:
            return self.array.shape[0]
        except AttributeError:
            return 0

    @property
    def bits(self):
        try:
            return self.array.shape[1]
        except AttributeError:
            return None

    def _check_fingerprints_are_valid(self, fprints):
        """Check if passed fingerprints fit database."""
        if fprints[0].level != self.level:
            raise ValueError("Provided fingerprints must have database level"
                             " {}".format(self.level))
        if self.fp_type is None:
            self.fp_type = fprints[0].__class__
        elif self.fp_type is not fprints[0].__class__:
            logging.warning("Database is of type {}. Fingerprints will be cast"
                            " to this type.")

    def __iter__(self):
        for i in range(self.fp_num):
            yield self.fp_type.from_vector(self.array[i, :], level=self.level,
                                           name=self.fp_names[i])

    def __add__(self, other):
        if self.level != other.level:
            raise TypeError("Cannot add databases with different levels")
        elif self.bits != other.bits:
            raise TypeError("Cannot add databases with different bit lengths")
        elif self.fp_type != other.fp_type:
            raise TypeError(
                "Cannot add databases with different fingerprint types")
        db = FingerprintDatabase(fp_type=self.fp_type, level=self.level)
        db.array = vstack([self.array, other.array]).tocsr()
        db.fp_names = self.fp_names + other.fp_names
        db.update_names_map()
        return db

    def __repr__(self):
        return "FingerprintDatabase(fp_type={}, level={}, name={})".format(
            self.fp_type, self.level, self.name)

    def __str__(self):
        return ("FingerprintDatabase[name: {}  fp_type: {}  level: {}"
                "  bits: {}  fp_num: {}]").format(self.name,
                                                  self.fp_type.__name__,
                                                  self.level, self.bits,
                                                  self.fp_num)

    def __len__(self):
        return self.fp_num

    def __getitem__(self, key):
        """Get list of fingerprints with name."""
        if isinstance(key, str):
            try:
                array = self.array[self.fp_names_to_indices[key]]
            except AttributeError:
                raise KeyError(
                    "fingerprint named {} is not in the database".format(key))
            return [self.fp_type.from_vector(array[i, :], level=self.level,
                                             name=key)
                    for i in range(array.shape[0])]
        elif isinstance(key, int):
            try:
                return self.fp_type.from_vector(self.array[key, :],
                                                level=self.level,
                                                name=self.fp_names[key])
            except (IndexError, AttributeError):
                raise IndexError("index out of range")
        else:
            raise TypeError("Key or index must be str or int.")

    def __getstate__(self):
        d = {}
        d["name"] = self.name
        d["fp_type"] = self.fp_type
        d["level"] = self.level
        d["array"] = self.array
        d["fp_names"] = self.fp_names
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__dict__["fp_names_to_indices"] = defaultdict(list)
        self.update_names_map()



"""Functions for various pipeline use cases.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
from .config.params import params_to_sections_dict
from .conformer.util import mol_from_smiles, mol_from_sdf, mol_to_sdf
from .conformer.generate import generate_conformers
from .fingerprint.generate import fprints_dict_from_mol


def params_to_dicts(params):
    """Get params dicts for pipeline functions from INI format params file."""
    sections_dict = params_to_sections_dict(params, auto=True)

    # preproc_params will eventually be returned separately, when there's a
    # pipeline function for protonation
    preproc_params = sections_dict.get("preprocessing", {})
    confgen_params = sections_dict.get("conformer_generation", {})
    confgen_params.update(preproc_params)
    fprint_params = sections_dict.get("fingerprinting", {})
    return confgen_params, fprint_params


def confs_from_smiles(smiles, name, confgen_params={}, save=False):
    """Generate conformations of molecule from SMILES string."""
    mol = mol_from_smiles(smiles, name)
    confgen_result = generate_conformers(mol, name, save=save,
                                         **confgen_params)
    mol = confgen_result[0]
    return mol


def sdf_from_smiles(smiles, name, confgen_params={}, out_file=None,
                    out_ext=".sdf.bz2"):
    """Generate conformations from SMILES string and save to SDF file."""
    mol = confs_from_smiles(smiles, name, confgen_params=confgen_params,
                            save=False)
    if out_file is None:
        out_file = name + out_ext
    mol_to_sdf(mol, out_file)


def fprints_from_fprints_dict(fprints_dict, level=-1):
    """Get fingerprint at `level` from dict of level to fingerprint."""
    fprints_list = fprints_dict.get(level,
                                    fprints_dict[max(fprints_dict.keys())])
    return fprints_list


def fprints_from_mol(mol, fprint_params={}, save=False):
    """Generate fingerprints for all `first` conformers in mol."""
    fprints_dict = fprints_dict_from_mol(mol, save=save, **fprint_params)
    level = fprint_params.get("level", -1)
    fprints_list = fprints_from_fprints_dict(fprints_dict, level=level)
    return fprints_list


def fprints_from_smiles(smiles, name, confgen_params={}, fprint_params={},
                        save=False):
    """Generate conformers and fingerprints from a SMILES string."""
    if save is False and "first" not in confgen_params:
        confgen_params["first"] = fprint_params.get("first", -1)
    mol = confs_from_smiles(smiles, name, confgen_params=confgen_params,
                            save=save)
    fprints_list = fprints_from_mol(mol, fprint_params=fprint_params,
                                    save=save)
    return fprints_list


def fprints_from_sdf(sdf_file, fprint_params={}, save=False):
    """Generate fingerprints from conformers in an SDF file."""
    mol = mol_from_sdf(sdf_file)
    fprints_list = fprints_from_mol(mol, fprint_params=fprint_params,
                                    save=save)
    return fprints_list



"""Integration tests for dependencies.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
import unittest


class RequiredDependenciesTestCase(unittest.TestCase):
    def test_rdkit(self):
        import rdkit

    def test_numpy(self):
        import numpy

    def test_scipy(self):
        import scipy

    def test_murmurhash(self):
        import mmh3

    def test_python_utilities(self):
        import python_utilities


class OptionalFeatureDependenciesTestCase(unittest.TestCase):
    def test_h5py(self):
        import h5py

    def test_standardiser(self):
        import standardiser

    def test_cxcalc(self):
        #Based on http://stackoverflow.com/a/11210902
        import subprocess
        import os
        try:
            devnull = open(os.devnull)
            subprocess.Popen(["cxcalc"], stdout=devnull,
                             stderr=devnull).communicate()
        except OSError as e:
            if e.errno == os.errno.ENOENT:
                raise


class OptionalParallelDependenciesTestCase(unittest.TestCase):
    def test_mpi4py(self):
        import mpi4py

    def test_concurrent(self):
        import concurrent.futures

    def test_python_utilities(self):
        import python_utilities.parallel


if __name__ == "__main__":
    unittest.main()



"""Generate conformers from SMILES or MOL2 files.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
from __future__ import division, print_function
import sys
import os
import logging
import argparse

from rdkit.Chem import AllChem
from python_utilities.parallel import Parallelizer, make_data_iterator, \
                                      ALL_PARALLEL_MODES
from python_utilities.scripting import setup_logging
from python_utilities.io_tools import touch_dir, HDF5Buffer
from e3fp.config.params import read_params, get_default_value, get_value
from e3fp.conformer.util import mol2_generator, smiles_generator, \
                                mol_from_mol2, mol_from_smiles, mol_to_sdf, \
                                mol_to_standardised_mol
from e3fp.conformer.generator import FORCEFIELD_CHOICES, ConformerGenerator

STANDARDISE_DEF = get_default_value("preprocessing", "standardise", bool)
NUM_CONF_DEF = get_default_value("conformer_generation", "num_conf", int)
FIRST_DEF = get_default_value("conformer_generation", "first", int)
POOL_MULTIPLIER_DEF = get_default_value("conformer_generation",
                                        "pool_multiplier", int)
RMSD_CUTOFF_DEF = get_default_value("conformer_generation", "rmsd_cutoff",
                                    float)
MAX_ENERGY_DIFF_DEF = get_default_value("conformer_generation",
                                        "max_energy_diff", float)
FORCEFIELD_DEF = get_default_value("conformer_generation", "forcefield")
OUTDIR_DEF = get_default_value("conformer_generation", "out_dir")
COMPRESS_DEF = get_default_value("conformer_generation", "compress")


def generate_conformers(input_mol, name=None, standardise=STANDARDISE_DEF,
                        num_conf=NUM_CONF_DEF, first=FIRST_DEF,
                        pool_multiplier=POOL_MULTIPLIER_DEF,
                        rmsd_cutoff=RMSD_CUTOFF_DEF,
                        max_energy_diff=MAX_ENERGY_DIFF_DEF,
                        forcefield=FORCEFIELD_DEF, params=None, out_file=None,
                        out_dir=OUTDIR_DEF, save=False, compress=COMPRESS_DEF,
                        overwrite=False):
    """Generate and save conformers for molecules.

    Parameters
    ----------
    input_mol : RDKit Mol
        Mol with a single conformer from which to generate conformers.
    name : str, optional
        Name of molecule.
    standardise : bool, optional
        Standardise mol before generating conformers.
    num_conf : int, optional
        If int, this is the target number of conformations. If -1, number
        of conformations is automatically chosen based on number of rotatable
        bonds.
    first : int, optional
        Number of first conformers to return. Does not impact conformer
        generator process, except may terminate conformer generation early when
        this many of conformers have been accepted.
    pool_multiplier : int, optional
        Factor to multiply by `num_conf`. The resulting number of conformations
        will be generated, then pruned to `num_conf`.
    rmsd_cutoff : float, optional
        RMSD threshold above which to accept two conformations as different
    max_energy_diff : float, optional
        Maximum energy difference between lowest energy conformer and any
        accepted conformer.
    forcefield : {'uff', 'mmff94', 'mmff94s'}, optional
        Forcefield to use for minimization of conformers.
    params : str or SafeConfigParser, optional
        Parameters or parameters file to use. If provided, all conformer
        generation options are overridden.
    out_file : str, optional
        Filename to save output, if `save` is True. If None, filename will be
        `name`.sdf, optionally with a compressed extension.
    out_dir : str, optional
        Directory where output files will be saved if `save` is True.
    save : bool, optional
        Save conformers to `out_file` in `out_dir`.
    compress : int, optional
        Compression of SDF files.
        None: auto. Mode is chosen based on extension, defaulting to SDF.
        0: File is not compressed.
        1: File is gzipped (.gz)
        2: File is bzipped (.bz2)
    overwrite : bool, optional
        Overwrite output files if they already exist.

    Returns
    -------
    bool
        If something went wrong, only return False. Otherwise return below.
    tuple
        Tuple with molecule name, number of rotatable bonds, numpy array of
        indices of final conformations, numpy array of energies of all
        conformations generated, and 2D numpy array of pairwise RMSDs between
        final conformations.
    """
    if name is None:
        name = input_mol.GetProp("_Name")

    if standardise:
        input_mol = mol_to_standardised_mol(input_mol)

    if save:
        if out_file is None:
            extensions = ("", ".gz", ".bz2")
            if compress not in (0, 1, 2):
                compress = 0
            out_file = os.path.join(out_dir,
                                    "%s.sdf%s" % (name, extensions[compress]))

        if os.path.exists(out_file) and not overwrite:
            logging.warning("%s already exists. Skipping." % (out_file))
            return False

    logging.info("Generating conformers for %s." % name)
    try:
        conf_gen = ConformerGenerator(num_conf=num_conf, first=first,
                                      pool_multiplier=pool_multiplier,
                                      rmsd_cutoff=rmsd_cutoff,
                                      max_energy_diff=max_energy_diff,
                                      forcefield=forcefield, get_values=True)
        mol, values = conf_gen.generate_conformers(input_mol)
        logging.info("Generated conformers for %s." % name)
    except:
        logging.warning("Problem generating conformers for %s." % name,
                        exc_info=True)
        return False

    if save:
        try:
            mol_to_sdf(mol, out_file)
            logging.info("Saved conformers for %s to %s." % (name, out_file))
        except:
            logging.warning(
                "Problem saving conformers for %s to %s." % (name, out_file),
                exc_info=True)
    return (mol, name, AllChem.CalcNumRotatableBonds(mol)) + values


def values_to_hdf5(hdf5_buffer, values):
    """Write specific values to `hdf5_buffer`.

    Parameters
    ----------
    hdf5_buffer : HDF5Buffer
        HDF5 buffer to write to.
    values : tuple
        Tuple of values to be written to buffer. Values and types should be
        `name` (str), `num_rotatable_bonds` (int), `target_conformers` (int),
        `indices` (numpy long array), `energies` (numpy float array),
        and `rmsd` (numpy float array). `rmsd` should be a square array with
        size along each dimension equal to length of `indices`.

    Returns
    -------
    bool
        True if success, False if not.
    """
    try:
        (name, num_rotatable_bonds,
         target_conformers, indices, energies, rmsd) = values

        group_dict = {"targetConfNum": {"data": target_conformers,
                                        "dtype": int},
                      "numRotatableBonds": {"data": num_rotatable_bonds,
                                            "dtype": int},
                      "indices": {"data": indices, "dtype": int,
                                  "compression": "gzip",
                                  "compression_opts": 9},
                      "energies": {"data": energies, "dtype": float,
                                   "compression": "gzip",
                                   "compression_opts": 9},
                      "rmsd": {"data": rmsd, "dtype": float,
                               "compression": "gzip",
                               "compression_opts": 9}
                      }

        hdf5_buffer.add_group(name, group_dict)
        logging.debug("Wrote values for %s to %s." % (name,
                                                      hdf5_buffer.filename))
        return True
    except:
        logging.error("Problem writing values to %s." % (
            hdf5_buffer.filename),
            exc_info=True)
        return False


def run(mol2=None, smiles=None, standardise=STANDARDISE_DEF,
        num_conf=NUM_CONF_DEF, first=FIRST_DEF,
        pool_multiplier=POOL_MULTIPLIER_DEF, rmsd_cutoff=RMSD_CUTOFF_DEF,
        max_energy_diff=MAX_ENERGY_DIFF_DEF, forcefield=FORCEFIELD_DEF,
        params=None, prioritize=False, out_dir=OUTDIR_DEF,
        compress=COMPRESS_DEF, overwrite=False, values_file=None, log=None,
        num_proc=None, parallel_mode=None, verbose=False):
    """Run conformer generation."""
    setup_logging(log, verbose=verbose)

    if params is not None:
        params = read_params(params)
        standardise = get_value(params, "preprocessing", "standardise", bool)
        num_conf = get_value(params, "conformer_generation", "num_conf", int)
        first = get_value(params, "conformer_generation", "first", int)
        pool_multiplier = get_value(params, "conformer_generation",
                                    "pool_multiplier", int)
        rmsd_cutoff = get_value(params, "conformer_generation", "rmsd_cutoff",
                                float)
        max_energy_diff = get_value(params, "conformer_generation",
                                    "max_energy_diff", float)
        forcefield = get_value(params, "conformer_generation", "forcefield")

    # check args
    if forcefield not in FORCEFIELD_CHOICES:
        raise ValueError(
            "Specified forcefield %s is not in valid options %r" % (
                forcefield, FORCEFIELD_CHOICES))

    para = Parallelizer(num_proc=num_proc, parallel_mode=parallel_mode)

    # Check to make sure args make sense
    if mol2 is None and smiles is None:
        if para.is_master():
            parser.print_usage()
            logging.error("Please provide mol2 file or a SMILES file.")
        sys.exit()

    if mol2 is not None and smiles is not None:
        if para.is_master():
            parser.print_usage()
            logging.error(
                "Please provide only a mol2 file OR a SMILES file.")
        sys.exit()

    if num_proc and num_proc < 1:
        if para.is_master():
            parser.print_usage()
            logging.error(
                "Please provide more than one processor with `--num_proc`.")
        sys.exit()

    # Set up input type
    if mol2 is not None:
        in_type = "mol2"
    elif smiles is not None:
        in_type = "smiles"

    if para.is_master():
        if in_type == "mol2":
            logging.info("Input type: mol2 file(s)")
            logging.info("Input file number: %d" % len(mol2))
            mol_iter = (mol_from_mol2(_mol2_file, _name,
                                      standardise=standardise)
                        for _mol2_file, _name in mol2_generator(*mol2))
        else:
            logging.info("Input type: Detected SMILES file(s)")
            logging.info("Input file number: %d" % len(smiles))
            mol_iter = (mol_from_smiles(_smiles, _name,
                                        standardise=standardise)
                        for _smiles, _name in smiles_generator(*smiles))

        if prioritize:
            logging.info("Prioritizing mols with low rotatable bond number and molecular weight first.")
            mols_with_properties = [
                (AllChem.CalcNumRotatableBonds(mol),
                 AllChem.CalcExactMolWt(mol), mol) for mol in mol_iter
                if mol is not None]
            data_iterator = make_data_iterator(
                (x[-1] for x in sorted(mols_with_properties)))
        else:
            data_iterator = make_data_iterator((x for x in mol_iter
                                                if x is not None))

        # Set up parallel-specific options
        logging.info("Parallel Type: %s" % para.parallel_mode)

        # Set other options
        touch_dir(out_dir)

        if not num_conf:
            num_conf = -1

        logging.info("Out Directory: %s" % out_dir)
        logging.info("Overwrite Existing Files: %s" % overwrite)
        if values_file is not None:
            if os.path.exists(values_file) and overwrite is not True:
                value_args = (values_file, 'a')
                logging.info("Values file: %s (append)" % (values_file))
            else:
                value_args = (values_file, 'w')
                logging.info("Values file: %s (new file)" % (values_file))
        if num_conf is None or num_conf == -1:
            logging.info("Target Conformer Number: auto")
        else:
            logging.info("Target Conformer Number: %d" % num_conf)
        if first is None or first == -1:
            logging.info("First Conformers Number: all")
        else:
            logging.info("First Conformers Number: %d" % first)
        logging.info("Pool Multiplier: %d" % pool_multiplier)
        logging.info("RMSD Cutoff: %.4g" % rmsd_cutoff)
        if max_energy_diff is None:
            logging.info("Maximum Energy Difference: None")
        else:
            logging.info(
                "Maximum Energy Difference: %.4g kcal" % (max_energy_diff))
        logging.info("Forcefield: %s" % forcefield.upper())

        logging.info("Starting.")
    else:
        data_iterator = iter([])

    gen_conf_kwargs = {"out_dir": out_dir, "params": params,
                       "num_conf": num_conf, "rmsd_cutoff": rmsd_cutoff,
                       "max_energy_diff": max_energy_diff,
                       "forcefield": forcefield,
                       "pool_multiplier": pool_multiplier, "first": first,
                       "save": True, "overwrite": overwrite,
                       "compress": compress}

    run_kwargs = {"kwargs": gen_conf_kwargs}

    results_iterator = para.run_gen(generate_conformers, data_iterator,
                                    **run_kwargs)

    if para.is_master() and values_file is not None:
        hdf5_buffer = HDF5Buffer(*value_args)

    for result, data in results_iterator:
        if para.is_master() and values_file is not None:
            values_to_hdf5(hdf5_buffer, result)

    if para.is_master() and values_file is not None:
        hdf5_buffer.flush()
        hdf5_buffer.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "Generate conformers from mol2 or SMILES",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-m', '--mol2', nargs='+', type=str, default=None,
                        help="Path to mol2 file(s), each with one molecule.")
    parser.add_argument('-s', '--smiles', nargs='+', type=str, default=None,
                        help="""Path to file(s) with SMILES and name. (space-
                             separated)""")
    parser.add_argument('--standardise', type=bool, default=STANDARDISE_DEF,
                        help="""Clean molecules before generating conformers
                             by standardisation.""")
    parser.add_argument('-n', '--num_conf', type=int, default=NUM_CONF_DEF,
                        help="""Set single number of conformers to use. -1
                             results in auto choosing.""")
    parser.add_argument('--first', type=int, default=FIRST_DEF,
                        help="""Set maximum number of first conformers to
                             accept. Conformer generation is unaffected,
                             except it may terminate early when this number of
                             conformers is reached.""")
    parser.add_argument('--pool_multiplier', type=int,
                        default=POOL_MULTIPLIER_DEF,
                        help="""Factor to multiply `num_conf` by to generate
                             conformers. Results are then pruned to
                             `num_conf`.""")
    parser.add_argument('-r', '--rmsd_cutoff', type=float,
                        default=RMSD_CUTOFF_DEF,
                        help="Choose RMSD cutoff between conformers")
    parser.add_argument('-e', '--max_energy_diff', type=float,
                        default=MAX_ENERGY_DIFF_DEF,
                        help="""Maximum energy difference between lowest
                             energy conformer and any accepted conformer.""")
    parser.add_argument('-f', '--forcefield', type=str,
                        choices=FORCEFIELD_CHOICES, default=FORCEFIELD_DEF,
                        help="""Choose forcefield for minimization.""")
    parser.add_argument('-o', '--out_dir', type=str, default=OUTDIR_DEF,
                        help="""Directory to save conformers.""")
    parser.add_argument('-C', '--compress', default=COMPRESS_DEF, type=int,
                        choices={None, 0, 1, 2},
                        help="""Compression to use for SDF files. None and 0
                             default to uncompressed ".sdf". 1 and 2 result in
                             gzipped and bzipped SDF files, respectively.""")
    parser.add_argument('-O', '--overwrite', action="store_true",
                        help="""Overwrite existing conformer files.""")
    parser.add_argument('--values_file', type=str, default=None,
                        help="""Save RMSDs and energies to specified hdf5
                             file.""")
    parser.add_argument('--prioritize', action="store_true",
                        help="""Prioritize likely fast molecules first.""")
    parser.add_argument('--params', type=str, default=None,
                        help="""INI formatted file with parameters. If
                             provided, all parameters controlling conformer
                             generation are ignored.""")
    parser.add_argument('-l', '--log', type=str, default=None,
                        help="Generate logfile.")
    parser.add_argument('-p', '--num_proc', type=int, default=None,
                        help="""Set number of processors to use.""")
    parser.add_argument('--parallel_mode', type=str, default=None,
                        choices=list(ALL_PARALLEL_MODES),
                        help="""Set number of processors to use.""")
    parser.add_argument('-v', '--verbose', action="store_true",
                        help="Run with extra verbosity.")
    params = parser.parse_args()

    kwargs = dict(params._get_kwargs())
    run(**kwargs)



"""Conformer generation.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
import logging

import numpy as np

from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import PropertyMol

# Heavily modified by Seth Axen from code under the following license
__author__ = "Steven Kearnes"
__copyright__ = "Copyright 2014, Stanford University"
__license__ = "3-clause BSD"

# options
FORCEFIELD_CHOICES = ('uff', 'mmff94', 'mmff94s')

# default values
NUM_CONF_DEF = -1
FIRST_DEF = -1
POOL_MULTIPLIER_DEF = 1
RMSD_CUTOFF_DEF = 0.5
MAX_ENERGY_DIFF_DEF = -1.
FORCEFIELD_DEF = 'uff'


class ConformerGenerator(object):

    """Generate conformers using RDKit.

    Procedure
    ---------
    1. Generate a pool of conformers.
    2. Minimize conformers.
    3. Filter conformers using an RMSD threshold and optional minimum energy
       difference.

    Note that pruning is done _after_ minimization, which differs from the
    protocol described in the references.

    References
    ----------
    * http://rdkit.org/docs/GettingStartedInPython.html
      #working-with-3d-molecules
    * http://pubs.acs.org/doi/full/10.1021/ci2004658
    * https://github.com/skearnes/rdkit-utils/blob/master/rdkit_utils/conformers.py
    """

    def __init__(self, num_conf=NUM_CONF_DEF, first=FIRST_DEF,
                 rmsd_cutoff=RMSD_CUTOFF_DEF,
                 max_energy_diff=MAX_ENERGY_DIFF_DEF,
                 forcefield=FORCEFIELD_DEF,
                 pool_multiplier=POOL_MULTIPLIER_DEF, get_values=False,
                 sparse_rmsd=True):
        """Initialize generator settings.

        Parameters
        ----------
        num_conf : int, optional
            Maximum number of conformers to generate (after pruning). -1
            results in auto selection of max_conformers.
        first : int, optional
            Terminate when this number of conformers has been accepted, and
            only return those conformers.
        pool_multiplier : int, optional
            Factor to multiply by max_conformers to generate the initial
            conformer pool. Since conformers are filtered after energy
            minimization, increasing the size of the pool increases the chance
            of identifying max_conformers unique conformers.
        rmsd_cutoff : float, optional
            RMSD cutoff for pruning conformers. If None or negative, no
            pruning is performed.
        max_energy_diff : float, optional
            If set, conformers with energies this amount above the minimum
            energy conformer are not accepted.
        forcefield : {'uff', 'mmff94', 'mmff94s'}, optional
            Force field to use for conformer energy calculation and
            minimization.
        get_values : boolean, optional
            Return tuple of key values, for storage.
        sparse_rmsd : bool, optional
            If `get_values` is True, instead of returning full symmetric RMSD
            matrix, only return flattened upper triangle.
        """
        self.max_conformers = num_conf
        self.first_conformers = first
        if not rmsd_cutoff or rmsd_cutoff < 0:
            rmsd_cutoff = -1.
        self.rmsd_cutoff = rmsd_cutoff

        if max_energy_diff is None or max_energy_diff < 0:
            max_energy_diff = -1.
        self.max_energy_diff = max_energy_diff

        if forcefield not in FORCEFIELD_CHOICES:
            raise ValueError(
                "%s is not a valid option for forcefield" % forcefield)
        self.forcefield = forcefield
        self.pool_multiplier = pool_multiplier
        self.get_values = get_values
        self.sparse_rmsd = sparse_rmsd

    def __call__(self, mol):
        """Generate conformers for a molecule.

        Parameters
        ----------
        mol : RDKit Mol
            Molecule.

        Returns
        ----------
        RDKit Mol : copy of the input molecule with embedded conformers
        """
        return self.generate_conformers(mol)

    def generate_conformers(self, mol):
        """Generate conformers for a molecule.

        Parameters
        ----------
        mol : RDKit Mol
            Molecule.

        Returns
        ----------
        RDKit Mol : copy of the input molecule with embedded conformers
        """
        # initial embedding
        mol = self.embed_molecule(mol)
        if not mol.GetNumConformers():
            msg = 'No conformers generated for molecule'
            if mol.HasProp('_Name'):
                name = mol.GetProp('_Name')
                msg += ' "{}".'.format(name)
            else:
                msg += '.'
            raise RuntimeError(msg)

        # minimization and filtering
        self.minimize_conformers(mol)
        mol, indices, energies, rmsds = self.filter_conformers(mol)

        if self.get_values is True:
            if self.sparse_rmsd:
                rmsds_mat = rmsds[np.triu_indices_from(rmsds, k=1)]
            else:
                rmsds_mat = rmsds
            return mol, (self.max_conformers, indices, energies, rmsds_mat)
        else:
            return mol

    @staticmethod
    def get_num_conformers(mol):
        """Return ideal number of conformers from rotatable bond number in model.

        Parameters
        ----------
        mol : Mol
            RDKit ``Mol`` object for molecule

        Yields
        ------
        num_conf : int
            Target number of conformers to accept
        """
        num_rot = AllChem.CalcNumRotatableBonds(mol)
        if num_rot < 8:
            return 50
        elif num_rot >= 8 and num_rot <= 12:
            return 200
        elif num_rot > 12:
            return 300
        else:
            return 0

    def embed_molecule(self, mol):
        """Generate conformers, possibly with pruning.

        Parameters
        ----------
        mol : RDKit Mol
            Molecule.
        """
        logging.debug("Adding hydrogens for %s" % mol.GetProp('_Name'))
        mol = Chem.AddHs(mol)  # add hydrogens
        logging.debug("Hydrogens added to %s" % mol.GetProp('_Name'))
        logging.debug("Sanitizing mol for %s" % mol.GetProp('_Name'))
        Chem.SanitizeMol(mol)
        logging.debug("Mol sanitized for %s" % mol.GetProp('_Name'))
        if (self.max_conformers == -1
                or type(self.max_conformers) is not int):
            self.max_conformers = self.get_num_conformers(mol)
        n_confs = self.max_conformers * self.pool_multiplier
        if self.first_conformers == -1:
            self.first_conformers = self.max_conformers
        logging.debug("Embedding %d conformers for %s" % (n_confs,
                                                          mol.GetProp('_Name'))
                      )
        AllChem.EmbedMultipleConfs(mol, numConfs=n_confs,
                                   maxAttempts=10 * n_confs,
                                   pruneRmsThresh=-1.,
                                   ignoreSmoothingFailures=True)
        logging.debug("Conformers embedded for %s" % mol.GetProp('_Name'))
        return mol

    def get_molecule_force_field(self, mol, conf_id=None, **kwargs):
        """Get a force field for a molecule.

        Parameters
        ----------
        mol : RDKit Mol
            Molecule.
        conf_id : int, optional
            ID of the conformer to associate with the force field.
        **kwargs : dict, optional
            Keyword arguments for force field constructor.
        """
        if self.forcefield == 'uff':
            ff = AllChem.UFFGetMoleculeForceField(
                mol, confId=conf_id, **kwargs)
        elif self.forcefield.startswith('mmff'):
            AllChem.MMFFSanitizeMolecule(mol)
            mmff_props = AllChem.MMFFGetMoleculeProperties(
                mol, mmffVariant=self.forcefield)
            ff = AllChem.MMFFGetMoleculeForceField(
                mol, mmff_props, confId=conf_id, **kwargs)
        else:
            raise ValueError("Invalid forcefield " +
                             "'{}'.".format(self.forcefield))
        return ff

    def minimize_conformers(self, mol):
        """Minimize molecule conformers.

        Parameters
        ----------
        mol : RDKit Mol
            Molecule.
        """
        logging.debug("Minimizing conformers for %s" % mol.GetProp('_Name'))
        for conf in mol.GetConformers():
            ff = self.get_molecule_force_field(mol, conf_id=conf.GetId())
            ff.Minimize()
        logging.debug("Conformers minimized for %s" % mol.GetProp('_Name'))

    def get_conformer_energies(self, mol):
        """Calculate conformer energies.

        Parameters
        ----------
        mol : RDKit Mol
            Molecule.

        Returns
        -------
        energies : array_like
            Minimized conformer energies.
        """
        num_conf = mol.GetNumConformers()
        energies = np.empty((num_conf,), dtype=np.float)
        for i, conf in enumerate(mol.GetConformers()):
            ff = self.get_molecule_force_field(mol, conf_id=conf.GetId())
            energies[i] = ff.CalcEnergy()
        return energies

    def filter_conformers(self, mol):
        """Filter conformers which do not meet an RMSD threshold.

        Parameters
        ----------
        mol : RDKit Mol
            Molecule.

        Returns
        -------
        A new RDKit Mol containing the chosen conformers, sorted by
        increasing energy.
        """
        logging.debug("Pruning conformers for %s" % mol.GetProp('_Name'))
        energies = self.get_conformer_energies(mol)
        energy_below_threshold = np.ones_like(energies, dtype=np.bool_)

        sort = np.argsort(energies)  # sort by increasing energy
        confs = np.array(mol.GetConformers())

        # remove hydrogens to speed up substruct match
        mol = Chem.RemoveHs(mol)
        accepted = []  # always accept lowest-energy conformer
        rejected = []
        rmsds = np.zeros((confs.shape[0], confs.shape[0]), dtype=np.float)
        for i, fit_ind in enumerate(sort):
            accepted_num = len(accepted)

            # always accept lowest-energy conformer
            if accepted_num == 0:
                accepted.append(fit_ind)

                # pre-compute if Es are in acceptable range of min E
                if self.max_energy_diff != -1.:
                    energy_below_threshold = (
                        energies <= energies[fit_ind] + self.max_energy_diff)

            # reject conformers after first_conformers is reached
            if accepted_num >= self.first_conformers:
                rejected.append(fit_ind)
                continue

            # check if energy is too high
            if not energy_below_threshold[fit_ind]:
                rejected.append(fit_ind)
                continue

            # get RMSD to selected conformers
            these_rmsds = np.zeros((accepted_num,), dtype=np.float)
            # reverse so all confs aligned to lowest energy
            for j, accepted_ind in self.reverse_enumerate(accepted):
                this_rmsd = AllChem.GetBestRMS(mol, mol,
                                               confs[accepted_ind].GetId(),
                                               confs[fit_ind].GetId())
                # reject conformers within the RMSD threshold
                if this_rmsd < self.rmsd_cutoff:
                    rejected.append(fit_ind)
                    break
                else:
                    these_rmsds[-j - 1] = this_rmsd
            else:
                rmsds[fit_ind, accepted] = these_rmsds
                rmsds[accepted, fit_ind] = these_rmsds
                accepted.append(fit_ind)

        # slice and order rmsds and energies to match accepted list
        rmsds = rmsds[np.ix_(accepted, accepted)]

        # create a new molecule with all conformers, sorted by energy
        new = PropertyMol.PropertyMol(mol)
        new.RemoveAllConformers()
        conf_ids = [conf.GetId() for conf in mol.GetConformers()]
        for i in accepted:
            conf = mol.GetConformer(conf_ids[i])
            new.AddConformer(conf, assignId=True)

        logging.debug("Conformers filtered for %s" % mol.GetProp('_Name'))
        return new, np.asarray(accepted, dtype=np.int), energies, rmsds

    @staticmethod
    def reverse_enumerate(iterable):
        """Enumerate, but with the last result first but still numbered last.

        Parameters
        ----------
        iterable : some 1-D iterable

        Returns
        -------
        iterable:
            Reverse of `enumerate` function
        """
        return zip(reversed(range(len(iterable))), reversed(iterable))

    # magic methods
    def __repr__(self):
        return """ConformerGenerator(num_conf=%r, first=%r,\
               \n                   pool_multiplier=%r, rmsd_cutoff=%r,\
               \n                   max_energy_diff=%r, forcefield=%r,\
               \n                   get_values=%r, sparse_rmsd=%r)""" % (
            self.max_conformers, self.first, self.pool_multiplier,
            self.rmsd_cutoff, self.max_energy_diff, self.forcefield,
            self.get_values, self.sparse_rmsd)

    def __str__(self):
        return self.__repr__()



"""Class for defining 3D atom environments.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
from __future__ import division, print_function
from functools import reduce

import numpy as np
import rdkit.Chem

from python_utilities.io_tools import smart_open
from e3fp.fingerprint import array_ops


PDB_LINE = ("HETATM{atom_id:>5d} {name:<4s} LIG A   1    "
            "{coord[0]:>8.3f}{coord[1]:>8.3f}{coord[2]:>8.3f}"
            "{occupancy:>6.2f}{temp:>6.2f}          {elem:>2s}{charge:>2s}")


class Shell(object):

    """A container for other Shells centered on an atom.

    Shells represent all atoms explicitly within a container. Atoms are
    represented by their ids. If atoms are provided instead of shells, they
    are converted to single-atom shells. A Substruct is generated from a Shell
    on the fly by recursion through member shells. An optional identifier may
    be set."""

    def __init__(self, center_atom, shells=set(), radius=None,
                 last_shell=None, identifier=None):
        if isinstance(center_atom, rdkit.Chem.Atom):
            center_atom = center_atom.GetIdx()
        elif not isinstance(center_atom, (int, np.integer)):
            raise TypeError("center_atom must be Atom or atom id")
        self._center_atom = center_atom

        self._shells = set()
        for shell in shells:
            if isinstance(shell, int):
                shell = Shell(shell)
            elif isinstance(shell, rdkit.Chem.Atom):
                shell = Shell(shell.GetIdx())
            elif not isinstance(shell, Shell):
                raise TypeError("shells must be Shells, Atoms, or atom ids")
            if shell.center_atom == self.center_atom:
                raise FormatError("member shells cannot be centered on same "
                                  "center_atom as new shell")
            self._shells.add(shell)
        self._shells = frozenset(self._shells)

        self.radius = radius
        self.last_shell = last_shell
        self.atoms = None
        self.substruct = None
        self.identifier = identifier
        self.is_duplicate = False
        self.duplicate = None

    @classmethod
    def from_substruct(cls, substruct):
        """Create shell with one shell for each atom in the substruct."""
        if substruct.center_atom is None:
            raise FormatError("Can only create Shell from Substruct if "
                              "center_atom is defined")
        atoms = substruct.atoms ^ {substruct.center_atom}
        return cls(substruct.center_atom, [Shell(x) for x in atoms])

    @property
    def center_atom(self):
        return self._center_atom

    @property
    def shells(self):
        return self._shells

    @property
    def atoms(self):
        """Get all atoms explicitly within the shell."""
        if self._atoms is None:
            self._atoms = set([self.center_atom, ])
            self._atoms.update([x.center_atom for x in self.shells])
        return self._atoms

    @atoms.setter
    def atoms(self, atoms):
        self._atoms = atoms

    @property
    def substruct(self):
        """Get substruct with all atoms implicitly within the shell."""
        if self._substruct is None:
            atom_sets = [set(x.substruct.atoms) for x in self.shells]
            if len(atom_sets) > 0:
                atoms = reduce(set.union, atom_sets)
            else:
                atoms = set()
            self._substruct = Substruct(
                center_atom=self.center_atom,
                atoms=atoms)
            self._substruct.shell = self
        return self._substruct

    @substruct.setter
    def substruct(self, substruct):
        if not isinstance(substruct, Substruct) and substruct is not None:
            raise TypeError("substruct must be of type Substruct")
        self._substruct = substruct

    def __repr__(self):
        return ('Shell(center_atom={!r}, shells={!r}, radius={!r}, '
                'last_shell={!r}, identifier={!r})').format(
                    self.center_atom, tuple(self.shells), self.radius,
                    self.last_shell, self.identifier)

    def __str__(self):
        return ('Shell(center_atom={!r}, atoms={!r}, radius={!r}, '
                'identifier={!r})').format(
                    self.center_atom, tuple(self.atoms), self.radius,
                    self.identifier)

    def __hash__(self):
        return hash((self.center_atom, self.shells))

    def __eq__(self, other):
        return ((self.center_atom == other.center_atom) and
                (self.shells == other.shells))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return 1 + len(self.shells)

    def __contains__(self, key):
        if isinstance(key, (int, rdkit.Chem.Atom)):
            key = Shell(key)
        return key in self.shells or key == self


class Substruct(object):

    """A container for atoms optionally centered on an atom.

    A Substruct represents all atoms implicitly within a Shell. Two Substructs
    are equal if they contain the same atoms."""

    def __init__(self, center_atom=None, atoms=set()):
        self.center_atom = center_atom
        self.shell = None
        self._atoms = set()
        for atom in atoms:
            if isinstance(atom, rdkit.Chem.Atom):
                atom = atom.GetIdx()
            elif not isinstance(atom, (int, np.integer)):
                raise TypeError("atoms must be Atom or atom id")
            self._atoms.add(atom)
        if self.center_atom is not None:
            self._atoms.add(self.center_atom)
        self._atoms = frozenset(self._atoms)
        self.transform_matrix = np.identity(4, dtype=np.float)

    @classmethod
    def from_shell(cls, shell):
        return shell.substruct

    @property
    def center_atom(self):
        return self._center_atom

    @center_atom.setter
    def center_atom(self, center_atom):
        if isinstance(center_atom, rdkit.Chem.Atom):
            center_atom = center_atom.GetIdx()
        elif (not isinstance(center_atom, (int, np.integer)) and
              center_atom is not None):
            raise TypeError("center_atom must be Atom or atom id")
        self._center_atom = center_atom

    @property
    def atoms(self):
        return self._atoms

    def __repr__(self):
        return "Substruct(center_atom={!r}, atoms={!r})".format(
            self.center_atom, tuple(self.atoms))

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return hash(self.atoms)

    def __eq__(self, other):
        return (self.atoms == other.atoms)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self.atoms)

    def __contains__(self, key):
        if isinstance(key, rdkit.Chem.Atom):
            key = key.GetIdx()
        return key in self.atoms


class FormatError(Exception):
    pass


# methods/classes for shell i/o
def shell_to_pdb(mol, shell, atom_coords, bound_atoms_dict, out_file=None,
                 reorient=True):
    """Append substructure within shell to PDB.

    Parameters
    ----------
    mol : RDKit Mol
        Input mol
    shell : Shell
        A shell
    atom_coords : dict
        Dict matching atom id to coordinates.
    bound_atoms_dict : dict
        Dict matching atom id to id of bound atoms.
    out_file : str or None, optional
        File to which to append coordinates.
    reorient : bool, optional
        Use the transformation matrix in the shell to align by the stereo
        quadrants. If no transformation matrix present, centers the center
        atom.

    Returns
    -------
    list of str: list of PDB file lines, if `out_file` not specified
    """
    remark = "REMARK 400"
    header_lines = [remark+" COMPOUND", remark+" "+mol.GetProp("_Name")]
    lines = header_lines + ["MODEL", ]
    atom_ids = sorted(shell.substruct.atoms)
    atoms = [mol.GetAtomWithIdx(x) for x in atom_ids]
    coords = np.asarray(list(map(atom_coords.get, atom_ids)), dtype=np.float64)
    if reorient:
        try:
            coords = array_ops.transform_array(shell.transform_matrix, coords)
        except AttributeError:
            coords -= atom_coords[shell.center_atom]

    for i, atom_id in enumerate(atom_ids):
        elem = atoms[i].GetSymbol()
        name = "{}{:d}".format(elem, atom_id + 1)
        charge = atoms[i].GetFormalCharge()
        if charge > 0:
            charge = "{:d}+".format(charge)
        elif charge < 0:
            charge = "{:d}-".format(abs(charge))
        else:
            charge = ""
        if atom_id == shell.center_atom:
            temp = 1.
        elif atom_id in shell.atoms:
            temp = .5
        else:
            temp = 0.
        pdb_entries = {"atom_id": atom_id,
                       "name": name,
                       "coord": coords[i, :].flatten(),
                       "occupancy": 0.,
                       "temp": temp,
                       "elem": elem,
                       "charge": charge}
        lines.append(PDB_LINE.format(**pdb_entries))

    # PLACEHOLDER FOR WRITING BONDS TO PDB
    # used_bonds = set()
    # write_bonds = []
    # for atom_id in atom_ids:
    #     write_bonds.append(atom_id)
    #     bound_atom_ids = bound_atoms_dict.get(atom_id, set())
    #     for bound_atom_id in bound_atom_ids:
    #         if (atom_id, bound_atom_id) in used_bonds:
    #             continue
    #         if len(write_bonds) > 3:
    #             lines.append("CONECT "+" ".join(map(str, write_bonds)))
    #             write_bonds = [atom_id,]
    #         write_bonds.append(bound_atom_id)
    #         used_bonds.add((atom_id, bound_atom_id))
    #         used_bonds.add((bound_atom_id, atom_id))

    #     lines.append("CONECT "+" ".join(map(str, write_bonds)))
    #     write_bonds = []

    lines.append("ENDMDL")

    if out_file is not None:
        with smart_open(out_file, "a") as f:
            for line in lines:
                f.write(line + "\n")
    else:
        return lines



"""Tests for Shell and Substruct objects.

Author: Seth Axen
E-mail: seth.axen@gmail.com
"""
import os
import unittest

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
PLANAR_SDF_FILE = os.path.join(DATA_DIR, "caffeine_planar.sdf.bz2")


class ShellCreationTestCases(unittest.TestCase):
    def test_error_when_center_not_atom(self):
        from e3fp.fingerprint.structs import Shell
        with self.assertRaises(TypeError):
            Shell(None)

    def test_error_when_shells_has_non_shell(self):
        from e3fp.fingerprint.structs import Shell
        atom = 0
        shells = [None]
        with self.assertRaises(TypeError):
            Shell(atom, shells)

    def test_creation_with_atoms_or_ids_equivalent(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        atom_ids = [x.GetIdx() for x in atoms]
        self.assertEqual(Shell(atoms[0], atoms[1:]),
                         Shell(atom_ids[0], atom_ids[1:]))

    def test_create_shell_no_shell(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        Shell(center_atom)

    def test_create_shell_with_same_center_fails(self):
        from e3fp.fingerprint.structs import Shell, FormatError
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        with self.assertRaises(FormatError):
            Shell(center_atom, atoms)

    def test_atoms_converted_to_shells(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        shell = Shell(center_atom, atoms[1:])
        for s in shell.shells:
            self.assertIsInstance(s, Shell)

    def test_creation_with_atoms_or_shells_equal(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        shells = list(map(Shell, atoms))
        center_atom = atoms[0]
        shell1 = Shell(center_atom, atoms[1:])
        shell2 = Shell(center_atom, shells[1:])
        self.assertEqual(shell1, shell2)

    def test_recursive_atom_shells_correct(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        shell1 = Shell(atoms[5], atoms[6:8])
        shell2 = Shell(atoms[2], atoms[3:5])
        shell = Shell(atoms[0], (shell1, shell2))
        self.assertEqual(shell.atoms,
                         {x.GetIdx() for x in (atoms[0], atoms[2], atoms[5])})


class ShellComparisonTestCases(unittest.TestCase):

    def test_shells_same_center_same_atoms_equal(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        shell1 = Shell(center_atom, atoms[1:])
        shell2 = Shell(center_atom, atoms[1:])
        self.assertEqual(shell1, shell2)

    def test_shells_diff_center_same_atoms_nonequal(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        shell1 = Shell(atoms[0], atoms[2:])
        shell2 = Shell(atoms[1], atoms[2:])
        self.assertNotEqual(shell1, shell2)

    def test_shells_same_center_diff_atoms_nonequal(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        shell1 = Shell(center_atom, atoms[1:])
        shell2 = Shell(center_atom, atoms[2:])
        self.assertNotEqual(shell1, shell2)

    def test_equal_shells_hash_to_same_value(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        shell1 = Shell(center_atom, atoms[1:])
        shell2 = Shell(center_atom, atoms[1:])
        self.assertEqual(hash(shell1), hash(shell2))

    def test_same_shell_hashes_to_same_value(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        shell = Shell(center_atom, atoms[1:])
        self.assertEqual(hash(shell), hash(shell))


class ShellSubstructInterfaceTestCases(unittest.TestCase):
    def test_recursive_shell_substruct_correct1(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        shell1 = Shell(atoms[5], atoms[6:8])
        shell2 = Shell(atoms[1], atoms[2:5])
        shell = Shell(atoms[0], (shell1, shell2))
        self.assertEqual(shell.substruct.atoms,
                         {x.GetIdx() for x in atoms[:8]})

    def test_recursive_shell_substruct_correct2(self):
        from e3fp.fingerprint.structs import Shell
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        shell1 = Shell(atoms[1], atoms[2:5])
        shell2 = Shell(atoms[5], {shell1})
        shell3 = Shell(atoms[6], atoms[7:10])
        shell4 = Shell(atoms[10], {shell3})
        shell = Shell(atoms[0], (shell2, shell4))
        self.assertEqual(shell.substruct.atoms,
                         {x.GetIdx() for x in atoms[:11]})

    def test_shell_creation_from_substruct_without_center_fails(self):
        from e3fp.fingerprint.structs import Shell, Substruct, FormatError
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        substruct = Substruct(None, atoms[:2])
        with self.assertRaises(FormatError):
            Shell.from_substruct(substruct)

    def test_shell_creation_from_substruct(self):
        from e3fp.fingerprint.structs import Shell, Substruct
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        substruct = Substruct(atoms[0], atoms[:2])
        shell = Shell.from_substruct(substruct)
        self.assertEqual(shell.atoms, substruct.atoms)

    def test_substruct_creation_from_shell(self):
        from e3fp.fingerprint.structs import Shell, Substruct
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        shell = Shell(atoms[0], atoms[1:])
        substruct = Substruct.from_shell(shell)
        self.assertEqual(shell.substruct, substruct)


class SubstructCreationTestCases(unittest.TestCase):
    def test_error_when_center_not_atom(self):
        from e3fp.fingerprint.structs import Substruct
        with self.assertRaises(TypeError):
            Substruct("foo")

    def test_error_when_atoms_has_non_atom(self):
        from e3fp.fingerprint.structs import Substruct
        atoms = [None]
        with self.assertRaises(TypeError):
            Substruct(atoms=atoms)

    def test_center_atom_auto_added_to_atoms(self):
        from e3fp.fingerprint.structs import Substruct
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        substruct = Substruct(center_atom, atoms[1:])
        self.assertIn(center_atom.GetIdx(), substruct.atoms)


class SubstructCreationComparisonCases(unittest.TestCase):

    def test_substructs_same_center_same_atoms_equal(self):
        from e3fp.fingerprint.structs import Substruct
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        substruct1 = Substruct(center_atom, atoms)
        substruct2 = Substruct(center_atom, atoms)
        self.assertEqual(substruct1, substruct2)

    def test_substructs_diff_center_same_atoms_equal(self):
        from e3fp.fingerprint.structs import Substruct
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        substruct1 = Substruct(atoms[0], atoms)
        substruct2 = Substruct(atoms[1], atoms)
        self.assertEqual(substruct1, substruct2)

    def test_substructs_same_center_diff_atoms_nonequal(self):
        from e3fp.fingerprint.structs import Substruct
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        substruct1 = Substruct(atoms[0], atoms[1:])
        substruct2 = Substruct(atoms[0], atoms[2:])
        self.assertNotEqual(substruct1, substruct2)

    def test_equal_shells_hash_to_same_value(self):
        from e3fp.fingerprint.structs import Substruct
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        substruct1 = Substruct(center_atom, atoms[1:])
        substruct2 = Substruct(center_atom, atoms[1:])
        self.assertEqual(hash(substruct1), hash(substruct2))

    def test_same_shells_hash_to_same_value(self):
        from e3fp.fingerprint.structs import Substruct
        from e3fp.conformer.util import mol_from_sdf
        mol = mol_from_sdf(PLANAR_SDF_FILE)
        atoms = list(mol.GetAtoms())
        center_atom = atoms[0]
        substruct = Substruct(center_atom, atoms[1:])
        self.assertEqual(hash(substruct), hash(substruct))


if __name__ == "__main__":
    unittest.main()
